# ruff: noqa: B006, C408, N802, N803, N806

import numpy as np


def adjlist_apply(
    X,
    W=None,
    alist=None,
    func=np.subtract,
    skip_verify=False,
    to_adjlist_kws=dict(drop_islands=None),
):
    """
    apply a function to an adajcency list, getting an adjacency list and result.

    Parameters
    ----------
    X       :   iterable
                an (N,P)-length iterable to apply ``func'' to. If (N,1), then `func`
                must take 2 arguments and return a single reduction. If P>1, then
                func must take two P-length arrays and return a single
                reduction of them.
    W       :   pysal.weights.W object
                a weights object that provides adjacency information
    alist   :   pandas DataFrame
                a table containing an adajacency list representation of a W matrix
    func    :   callable
                a function taking two arguments and returning a single argument.
                This will be evaluated for every (focal, neighbor) pair, or each
                row of the adjacency list. If `X` has more than one column, this
                function should take two arrays and provide a single scalar in
                return.
                Example scalars include: lambda x,y: x < y, np.subtract
                Example multivariates: lambda (x,y): np.all(x < y)''
                                       lambda (x,y): np.sum((x-y)**2)
                                       sklearn.metrics.euclidean_distance
    skip_verify: bool
                Whether or not to skip verifying that the
                W is the same as an adjacency list.
                Do this if you are certain the adjacency
                list and W agree and would like to
                avoid re-instantiating a W from the adjacency list.
    to_adjlist_kws : dict
        Keyword arguments for ``W.to_adjlist()``.
        Default is ``dict(drop_islands=None)``.

    Returns
    -------
    an adjacency list (or modifies alist inplace) with the function applied to each row.
    """
    try:
        import pandas as pd
    except ImportError:
        raise ImportError("pandas must be installed to use this function") from None
    w, alist = _get_W_and_alist(W, alist, to_adjlist_kws, skip_verify=skip_verify)
    if len(X.shape) > 1:
        if X.shape[-1] > 1:
            return _adjlist_mvapply(
                X,
                W=w,
                alist=alist,
                func=func,
                skip_verify=skip_verify,
                to_adjlist_kws=to_adjlist_kws,
            )
    else:
        vec = np.asarray(X).flatten()
    ids = np.asarray(w.id_order)[:, None]
    table = pd.DataFrame(ids, columns=["id"])
    table = pd.concat((table, pd.DataFrame(vec[:, None], columns=("att",))), axis=1)
    alist_atts = pd.merge(alist, table, how="left", left_on="focal", right_on="id")
    alist_atts = pd.merge(
        alist_atts,
        table,
        how="left",
        left_on="neighbor",
        right_on="id",
        suffixes=("_focal", "_neighbor"),
    )
    alist_atts.drop(["id_focal", "id_neighbor"], axis=1, inplace=True)
    alist_atts[func.__name__] = alist_atts[["att_focal", "att_neighbor"]].apply(
        lambda x: func(x.att_focal, x.att_neighbor), axis=1
    )
    return alist_atts


def _adjlist_mvapply(
    X, W=None, alist=None, func=None, skip_verify=False, to_adjlist_kws=dict()
):
    try:
        import pandas as pd
    except ImportError:
        raise ImportError("pandas must be installed to use this function") from None
    assert len(X.shape) == 2, "data is not two-dimensional"
    w, alist = _get_W_and_alist(W, alist, to_adjlist_kws, skip_verify=skip_verify)
    assert X.shape[0] == w.n, "number of samples in X does not match W"
    try:
        names = X.columns.tolist()
    except AttributeError:
        names = list(map(str, list(range(X.shape[1]))))
    ids = np.asarray(w.id_order)[:, None]
    table = pd.DataFrame(ids, columns=["id"])
    table = pd.concat((table, pd.DataFrame(X, columns=names)), axis=1)
    alist_atts = pd.merge(alist, table, how="left", left_on="focal", right_on="id")
    alist_atts = pd.merge(
        alist_atts,
        table,
        how="left",
        left_on="neighbor",
        right_on="id",
        suffixes=("_focal", "_neighbor"),
    )
    alist_atts.drop(["id_focal", "id_neighbor"], axis=1, inplace=True)
    alist_atts[func.__name__] = list(
        map(
            func,
            list(
                zip(
                    alist_atts.filter(like="_focal").values,
                    alist_atts.filter(like="_neighbor").values,
                    strict=True,
                )
            ),
        )
    )
    return alist_atts


def _get_W_and_alist(W, alist, to_adjlist_kws, skip_verify=False):
    """
    Either:
    1. compute a W from an alist
    2. adjacencylist from a W
    3. raise ValueError if neither are provided,
    4. raise AssertionError if both W and adjlist are provided and don't match.
    If this completes successfully, the W/adjlist will both be
    returned and are checked for equality.
    """
    if (alist is None) and (W is not None):
        alist = W.to_adjlist(**to_adjlist_kws)
    elif (W is None) and (alist is not None):
        from .weights import W

        W = W.from_adjlist(alist, **to_adjlist_kws)
    elif (W is None) and (alist is None):
        raise ValueError("Either W or Adjacency List must be provided")
    elif (W is not None) and (alist is not None) and (not skip_verify):
        from .weights import W as W_

        np.testing.assert_allclose(
            W.sparse.toarray(), W_.from_adjlist(alist).sparse.toarray()
        )
    return W, alist


def adjlist_map(
    data,
    funcs=(np.subtract,),
    W=None,
    alist=None,
    focal_col="focal",  # noqa: ARG001
    neighbor_col="neighbor",  # noqa: ARG001
    to_adjlist_kws=dict(drop_islands=None),
):
    """
    Map a set of functions over a W or adjacency list

    Parameters
    ----------
    data        :   np.ndarray or pandas dataframe
                    N x P array of N observations and P covariates.
    funcs       :   iterable or callable
                    a function to apply to each of the P columns in ``data'', or a
                    list of functions to apply to each column of P. This function
                    must take two arguments, compare them, and return a value. Examples
                    may be ``lambda x,y: x < y'' or ``np.subtract''.
    W           :   pysal.weights.W object
                    a pysal weights object. If not provided, one is constructed from
                    the given adjacency list.
    alist       :   pandas dataframe
                    an adjacency list representation of a weights matrix. If not
                    provided, one is constructed from the weights object. If both are
                    provided, they are validated against one another to ensure they
                    provide identical weights matrices.
    focal_col   :   string
                    name of column in alist containing the focal observation ids
    neighbor_col:   string
                    name of column in alist containing the neighboring observation ids
    to_adjlist_kws : dict
        Keyword arguments for ``W.to_adjlist()``.
        Default is ``dict(drop_islands=None)``.

    Returns
    -------
    returns an adjacency list (or modifies one if provided)
    with each function applied to the column of the data.
    """
    try:
        import pandas as pd
    except ImportError:
        raise ImportError("pandas must be installed to use this function") from None
    if isinstance(data, pd.DataFrame):
        names = data.columns
        data = data.values
    else:
        names = [str(i) for i in range(data.shape[1])]
    assert data.shape[0] == W.n, "shape of data does not match shape of adjacency"
    if callable(funcs):
        funcs = (funcs,)
    if len(funcs) == 1:
        funcs = [funcs[0] for _ in range(data.shape[1])]
    assert data.shape[1] == len(
        funcs
    ), "shape of data does not match the number of functions provided"
    w, alist = _get_W_and_alist(W, alist, to_adjlist_kws)
    fnames = {f.__name__ for f in funcs}
    for i, (column, function) in enumerate(zip(data.T, funcs, strict=True)):
        alist = adjlist_apply(
            column, W=W, alist=alist, skip_verify=True, to_adjlist_kws=to_adjlist_kws
        )
        alist.drop(["att_focal", "att_neighbor"], axis=1, inplace=True)
        alist = alist.rename(
            columns={function.__name__: "_".join((function.__name__, names[i]))}
        )
        fnames.update((function.__name__,))
    return alist


def filter_adjlist(adjlist, focal_col="focal", neighbor_col="neighbor"):
    """
    This dedupes an adjacency list by examining both (a,b)
    and (b,a) when (a,b) is enountered. The removal is done
    in order of the iteration order of the input adjacency list.
    So, if a special order of removal is desired, you need to
    sort the list before this function.

    Parameters
    ----------
    adjlist     :   pandas DataFrame
                    a dataframe that contains focal and neighbor columns
    focal_col   :   string
                    the name of the column with the focal observation id
    neighbor_col:   string
                    the name of the column with the neighbor observation id

    Returns
    -------
    an adjacency table with reversible entries removed.
    """
    edges = adjlist.loc[:, [focal_col, neighbor_col]]
    undirected = set()
    to_remove = []
    for index, *edge in edges.itertuples(name=None):
        edge = tuple(edge)
        if edge in undirected or edge[::-1] in undirected:
            to_remove.append(index)
        else:
            undirected.add(edge)
            undirected.add(edge[::-1])
    adjlist = adjlist.drop(to_remove)
    return adjlist



# ruff: noqa: SIM115

from ...weights import W
from .. import fileio

__author__ = "Myunghwa Hwang <mhwang4@gmail.com>"
__all__ = ["GeoBUGSTextIO"]


class GeoBUGSTextIO(fileio.FileIO):
    """Opens, reads, and writes weights file objects in the text format
    used in `GeoBUGS <http://www.openbugs.net/Manuals/GeoBUGS/Manual.html>`_.
    `GeoBUGS` generates a spatial weights matrix as an R object and writes
    it out as an ASCII text representation of the R object. An exemplary
    `GeoBUGS` text file is as follows.

    ```
    list([CARD], [ADJ], [WGT], [SUMNUMNEIGH])
    ```

    where ``[CARD]`` and ``[ADJ]`` are required but the others are optional.
    PySAL assumes ``[CARD]`` and ``[ADJ]`` always exist in an input text file.
    It can read a `GeoBUGS` text file, even when its content is not written
    in the order of ``[CARD]``, ``[ADJ]``, ``[WGT]``, and ``[SUMNUMNEIGH]``.
    It always writes all of ``[CARD]``, ``[ADJ]``, ``[WGT]``, and ``[SUMNUMNEIGH]``.
    PySAL does not apply text wrapping during file writing.

    In the above example,

    ```
    [CARD]:
        num = c([a list of comma-splitted neighbor cardinalities])

    [ADJ]:
        adj = c ([a list of comma-splitted neighbor IDs])

        If caridnality is zero, neighbor IDs are skipped. The ordering of
        observations is the same in both ``[CARD]`` and ``[ADJ]``.
        Neighbor IDs are record numbers starting from one.

    [WGT]:
        weights = c([a list of comma-splitted weights])
        The restrictions for [ADJ] also apply to ``[WGT]``.

    [SUMNUMNEIGH]:
        sumNumNeigh = [The total number of neighbor pairs]
        the total number of neighbor pairs  is an integer
        value and the same as the sum of neighbor cardinalities.
    ```

    Notes
    -----

    For the files generated from R the ``spdep``, ``nb2WB``, and ``dput``
    functions. It is assumed that the value for the control parameter of
    the ``dput`` function is ``NULL``. Please refer to R ``spdep`` and
    ``nb2WB`` functions help files.

    References
    ----------

    * **Thomas, A., Best, N., Lunn, D., Arnold, R., and Spiegelhalter, D.**
        (2004) GeoBUGS User Manual. R spdep nb2WB function help file.

    """

    FORMATS = ["geobugs_text"]
    MODES = ["r", "w"]

    def __init__(self, *args, **kwargs):
        args = args[:2]
        fileio.FileIO.__init__(self, *args, **kwargs)
        self.file = open(self.dataPath, self.mode)

    def read(self, n=-1):  # noqa: ARG002
        """Read a GeoBUGS text file.

        Returns
        -------
        w : libpysal.weights.W
            A PySAL `W` object.

        Examples
        --------

        Type ``dir(w)`` at the interpreter to see what methods are supported.
        Open a `GeoBUGS` text file and read it into a PySAL weights object.

        >>> import libpysal
        >>> w = libpysal.io.open(
        ...     libpysal.examples.get_path('geobugs_scot'), 'r', 'geobugs_text'
        ... ).read()

        Get the number of observations from the header.

        >>> w.n
        56

        Get the mean number of neighbors.

        >>> w.mean_neighbors
        4.178571428571429

        Get neighbor distances for a single observation.

        >>> w[1] == dict({9: 1.0, 19: 1.0, 5: 1.0})
        True

        """

        self._complain_ifclosed(self.closed)

        w = self._read()

        return w

    def seek(self, pos) -> int:
        if pos == 0:
            self.file.seek(0)
            self.pos = 0

    def _read(self):
        """Reads in a `GeoBUGSTextIO` object.

        Raises
        ------
        StopIteration
            Raised at the EOF.

        Returns
        -------
        w : libpysal.weights.W
            A PySAL `W` object.

        """

        if self.pos > 0:
            raise StopIteration

        fbody = self.file.read()
        body_structure = {}

        for i in ["num", "adj", "weights", "sumNumNeigh"]:
            i_loc = fbody.find(i)

            if i_loc != -1:
                body_structure[i] = (i_loc, i)

        body_sequence = sorted(body_structure.values())
        body_sequence.append((-1, "eof"))

        for i in range(len(body_sequence) - 1):
            part, next_part = body_sequence[i], body_sequence[i + 1]
            start, end = part[0], next_part[0]
            part_text = fbody[start:end]

            part_length, start, end = len(part_text), 0, -1

            for c in range(part_length):
                if part_text[c].isdigit():
                    start = c
                    break

            for c in range(part_length - 1, 0, -1):
                if part_text[c].isdigit():
                    end = c + 1
                    break

            part_text = part_text[start:end]
            part_text = part_text.replace("\n", "")
            value_type = int

            if part[1] == "weights":
                value_type = float

            body_structure[part[1]] = [value_type(v) for v in part_text.split(",")]

        cardinalities = body_structure["num"]
        adjacency = body_structure["adj"]
        raw_weights = [1.0] * int(sum(cardinalities))

        if "weights" in body_structure and isinstance(body_structure["weights"], list):
            raw_weights = body_structure["weights"]

        no_obs = len(cardinalities)
        neighbors = {}
        weights = {}
        pos = 0

        for i in range(no_obs):
            neighbors[i + 1] = []
            weights[i + 1] = []
            no_nghs = cardinalities[i]

            if no_nghs > 0:
                neighbors[i + 1] = adjacency[pos : pos + no_nghs]
                weights[i + 1] = raw_weights[pos : pos + no_nghs]

            pos += no_nghs

        self.pos += 1

        w = W(neighbors, weights)

        return w

    def write(self, obj):
        """Writes a weights object to the opened text file.

        Parameters
        ----------
        obj : libpysal.weights.W
            A PySAL `W` object.

        Raises
        ------
        TypeError
            Raised when the input ``obj`` is not a PySAL `W`.

        Examples
        --------

        >>> import tempfile, libpysal, os
        >>> testfile = libpysal.io.open(
        ...     libpysal.examples.get_path('geobugs_scot'), 'r', 'geobugs_text'
        ... )
        >>> w = testfile.read()

        Create a temporary file for this example.

        >>> f = tempfile.NamedTemporaryFile(suffix='')

        Reassign to a new variable.

        >>> fname = f.name

        Close the temporary named file.

        >>> f.close()

        Open the new file in write mode.

        >>> o = libpysal.io.open(fname, 'w', 'geobugs_text')

        Write the Weights object into the open file.

        >>> o.write(w)
        >>> o.close()

        Read in the newly created text file.

        >>> wnew =  libpysal.io.open(fname, 'r', 'geobugs_text').read()

        Compare values from old to new.

        >>> wnew.pct_nonzero == w.pct_nonzero
        True

        Clean up the temporary file created for this example.

        >>> os.remove(fname)

        """

        self._complain_ifclosed(self.closed)

        if issubclass(type(obj), W):
            cardinalities, neighbors, weights = [], [], []
            for i in obj.id_order:
                cardinalities.append(obj.cardinalities[i])
                neighbors.extend(obj.neighbors[i])
                weights.extend(obj.weights[i])

            self.file.write("list(")
            self.file.write("num=c(%s)," % ",".join(map(str, cardinalities)))
            self.file.write("adj=c(%s)," % ",".join(map(str, neighbors)))
            self.file.write("sumNumNeigh=%i)" % sum(cardinalities))
            self.pos += 1

        else:
            raise TypeError("Expected a PySAL weights object, got: %s." % (type(obj)))

    def close(self):
        self.file.close()
        fileio.FileIO.close(self)



import warnings
from itertools import permutations

import geopandas
import numpy as np
import pandas as pd
import shapely
from packaging.version import Version

GPD_013 = Version(geopandas.__version__) >= Version("0.13")
PANDAS_GE_21 = Version(pd.__version__) >= Version("2.1.0")


def _sparse_to_arrays(sparray, ids=None):
    """Convert sparse array to arrays of adjacency"""
    sparray = sparray.tocoo(copy=False)
    if ids is not None:
        ids = np.asarray(ids)
        if sparray.shape[0] != ids.shape[0]:
            raise ValueError(
                f"The length of ids ({ids.shape[0]}) does not match "
                f"the shape of sparse {sparray.shape}."
            )

        sorter = sparray.row.argsort()
        head = ids[sparray.row][sorter]
        tail = ids[sparray.col][sorter]
        data = sparray.data[sorter]
    else:
        sorter = sparray.row.argsort()
        head = sparray.row[sorter]
        tail = sparray.col[sorter]
        data = sparray.data[sorter]
        ids = np.arange(sparray.shape[0], dtype=int)
    return _resolve_islands(head, tail, ids, data)


def _jitter_geoms(coordinates, geoms, seed=None):
    """
    Jitter geometries based on the smallest required movements to induce
    uniqueness. For each point, this samples a radius and angle uniformly
    at random from the unit circle, rescales it to a circle of values that
    are extremely small relative to the precision of the input, and
    then displaces the point. For a non-euclidean geometry, like latitude
    longitude coordinates, this will distort according to a plateé carree
    projection, jittering slightly more in the x direction than the y direction.
    """
    rng = np.random.default_rng(seed=seed)
    dtype = coordinates.dtype
    if dtype not in (np.float32, np.float64):
        # jittering requires us to cast ints to float
        # and the rng.random generator only works with float32 and float64
        dtype = np.float32
    # the resolution is the approximate difference between two floats
    # that can be resolved at the given dtype.
    resolution = np.finfo(dtype).resolution
    r = rng.random(size=coordinates.shape[0], dtype=dtype) ** 0.5 * resolution
    theta = rng.random(size=coordinates.shape[0], dtype=dtype) * np.pi * 2
    # converting from polar to cartesian
    dx = r + np.sin(theta)
    dy = r + np.cos(theta)
    # then adding the displacements
    coordinates = coordinates + np.column_stack((dx, dy))
    geoms = geopandas.GeoSeries(geopandas.points_from_xy(*coordinates.T, crs=geoms.crs))
    return coordinates, geoms


def _induce_cliques(adjtable, clique_to_members, fill_value=1):
    """
    induce cliques into the input graph. This connects everything within a
    clique together, as well as connecting all things outside of the clique
    to all members of the clique.

    This does not guarantee/understand ordering of the *output* adjacency table.
    """
    adj_across_clique = (
        adjtable.merge(
            clique_to_members["input_index"], left_on="focal", right_index=True
        )
        .explode("input_index")
        .rename(columns={"input_index": "subclique_focal"})
        .merge(clique_to_members["input_index"], left_on="neighbor", right_index=True)
        .explode("input_index")
        .rename(columns={"input_index": "subclique_neighbor"})
        .reset_index()
        .drop(["focal", "neighbor", "index"], axis=1)
        .rename(columns={"subclique_focal": "focal", "subclique_neighbor": "neighbor"})
    )[["focal", "neighbor", "weight"]]
    is_multimember_clique = clique_to_members["input_index"].str.len() > 1
    adj_within_clique = (
        clique_to_members[is_multimember_clique]["input_index"]
        .apply(lambda x: list(permutations(x, 2)))
        .explode()
        .apply(pd.Series)
        .rename(columns={0: "focal", 1: "neighbor"})
        .assign(weight=fill_value)
    )

    new_adj = pd.concat(
        (adj_across_clique, adj_within_clique), ignore_index=True, axis=0
    ).reset_index(drop=True)

    return new_adj


def _neighbor_dict_to_edges(neighbors, weights=None):
    """
    Convert a neighbor dict to a set of (head, tail, weight) edges, assuming
    that the any self-loops have a weight of zero.
    """
    idxs = pd.Series(neighbors).explode()
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            "Downcasting object dtype arrays on .fillna, .ffill, .bfill ",
            FutureWarning,
        )
        idxs = idxs.fillna(pd.Series(idxs.index, index=idxs.index))  # self-loops
    heads, tails = idxs.index.values, idxs.values
    tails = tails.astype(heads.dtype)
    if weights is not None:
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                "Downcasting object dtype arrays on .fillna, .ffill, .bfill ",
                FutureWarning,
            )
            data_array = pd.Series(weights).explode().fillna(0).values
        if not pd.api.types.is_numeric_dtype(data_array):
            data_array = pd.to_numeric(data_array)
    else:
        data_array = np.ones(idxs.shape[0], dtype=int)
        data_array[heads == tails] = 0
    return heads, tails, data_array


def _build_coincidence_lookup(geoms):
    """
    Identify coincident points and create a look-up table for the coincident geometries.
    """
    valid_coincident_geom_types = set(("Point",))  # noqa: C405
    if not set(geoms.geom_type) <= valid_coincident_geom_types:
        raise ValueError(
            "coindicence checks are only well-defined for "
            f"geom_types: {valid_coincident_geom_types}"
        )
    max_coincident = geoms.geometry.duplicated().sum()
    if GPD_013:
        lut = (
            geoms.to_frame("geometry")
            .reset_index()
            .groupby("geometry")["index"]
            .agg(list)
            .reset_index()
        )
    else:
        lut = (
            geoms.to_wkb()
            .to_frame("geometry")
            .reset_index()
            .groupby("geometry")["index"]
            .agg(list)
            .reset_index()
        )
        lut["geometry"] = geopandas.GeoSeries.from_wkb(lut["geometry"])

    lut = geopandas.GeoDataFrame(lut)
    return max_coincident, lut.rename(columns={"index": "input_index"})


def _validate_geometry_input(geoms, ids=None, valid_geometry_types=None):
    """
    Ensure that input geometries are always aligned to (and refer back to)
    inputted geometries. Geoms can be a GeoSeries, GeoDataFrame, numpy.array
    with a geometry dtype, or a point array.

    is will always align to geoms.

    the returned coordinates will always pertain to geoms, but may be
    longer than geoms (such as when geoms represents polygons).
    """
    if isinstance(geoms, geopandas.GeoSeries | geopandas.GeoDataFrame):
        geoms = geoms.geometry
        if ids is None:
            ids = geoms.index
        ids = np.asarray(ids)
        geom_types = set(geoms.geom_type)
        if valid_geometry_types is not None:
            if isinstance(valid_geometry_types, str):
                valid_geometry_types = (valid_geometry_types,)
            valid_geometry_types = set(valid_geometry_types)
            if not geom_types <= valid_geometry_types:
                raise ValueError(
                    "This Graph type is only well-defined for "
                    f"geom_types: {valid_geometry_types}."
                )
        coordinates = shapely.get_coordinates(geoms)
        geoms = geoms.copy()
        geoms.index = ids
        return coordinates, ids, geoms
    elif isinstance(geoms.dtype, geopandas.array.GeometryDtype):
        return _validate_geometry_input(
            geopandas.GeoSeries(geoms),
            ids=ids,
            valid_geometry_types=valid_geometry_types,
        )
    else:
        if (geoms.ndim == 2) and (geoms.shape[1] == 2):
            return _validate_geometry_input(
                geopandas.points_from_xy(*geoms.T),
                ids=ids,
                valid_geometry_types=valid_geometry_types,
            )
    raise ValueError(
        "input geometry type is not supported. Input must either be a "
        "geopandas.GeoSeries, geopandas.GeoDataFrame, a numpy array with a geometry "
        "dtype, or an array of coordinates."
    )


def _validate_sparse_input(sparse, ids=None):
    assert (
        sparse.shape[0] == sparse.shape[1]
    ), "coordinates should represent a distance matrix if metric='precomputed'"
    return _sparse_to_arrays(sparse, ids)


def _vec_euclidean_distances(x_vec, y_vec):
    """
    compute the euclidean distances along corresponding rows of two arrays
    """
    return ((x_vec - y_vec) ** 2).sum(axis=1) ** 0.5


def _evaluate_index(data):
    """Helper to get ids from any input."""
    if isinstance(data, pd.Series | pd.DataFrame):
        return data.index
    elif hasattr(data, "shape"):
        return pd.RangeIndex(0, data.shape[0])
    else:
        return pd.RangeIndex(0, len(data))


def _resolve_islands(heads, tails, ids, weights):
    """
    Induce self-loops for a collection of ids and links describing a
    contiguity graph. Induced self-loops will have zero weight.
    """
    islands = pd.Index(ids).difference(pd.Index(heads))
    if islands.shape != (0,):
        heads = np.hstack((heads, islands))
        tails = np.hstack((tails, islands))
        weights = np.hstack((weights, np.zeros_like(islands, dtype=int)))

    # ensure proper order after adding isolates to the end
    adjacency = pd.Series(weights, index=pd.MultiIndex.from_arrays([heads, tails]))
    adjacency = adjacency.reindex(ids, level=0).reindex(ids, level=1)
    return (
        adjacency.index.get_level_values(0),
        adjacency.index.get_level_values(1),
        adjacency.values,
    )


def _reorder_adjtable_by_ids(adjtable, ids):
    if PANDAS_GE_21:
        # ensure proper sorting
        sorted_index = (
            adjtable[["focal", "neighbor"]]
            .map(list(ids).index)
            .sort_values(["focal", "neighbor"])
            .index
        )
    else:
        # ensure proper sorting
        sorted_index = (
            adjtable[["focal", "neighbor"]]
            .applymap(list(ids).index)
            .sort_values(["focal", "neighbor"])
            .index
        )
    return adjtable.iloc[sorted_index].reset_index(drop=True)



# ruff: noqa: N802, N806, N815, SIM115

from struct import pack, unpack

import numpy as np

from ...weights import W
from ...weights.util import remap_ids
from .. import fileio

__author__ = "Myunghwa Hwang <mhwang4@gmail.com>"
__all__ = ["ArcGISSwmIO"]


class ArcGISSwmIO(fileio.FileIO):
    """Opens, reads, and writes weights file objects in ArcGIS ``.swm`` format.
    Spatial weights objects in the ArcGIS ``.swm`` format are used in ArcGIS
    Spatial Statistics tools. Particularly, this format can be directly used
    with the tools under the category of Mapping Clusters.

    The values for``ORG_i`` and ``DST_i`` should be integers, as ArcGIS Spatial
    Statistics tools support only unique integer IDs. For the case where a
    weights object uses non-integer IDs, `ArcGISSwmIO` allows users to use
    internal IDs corresponding to record numbers, instead of original IDs.
    The specifics of each part of the above structure is as follows.

    .. table:: ArcGIS SWM Components
    ============ ============ ==================================== ================================
        Part      Data type           Description                               Length
    ============ ============ ==================================== ================================
     ID_VAR_NAME  ASCII TEXT  ID variable name                     Flexible (Up to the 1st ;)
     ESRI_SRS     ASCII TEXT  ESRI spatial reference system        Flexible (Btw the 1st ; and \\n)
     NO_OBS       l.e. int    Number of observations               4
     ROW_STD      l.e. int    Whether or not row-standardized      4
     WGT_i
     ORG_i        l.e. int    ID of observaiton i                  4
     NO_NGH_i     l.e. int    Number of neighbors for obs. i (m)   4
     NGHS_i
     DSTS_i       l.e. int    IDs of all neighbors of obs. i       4*m
     WS_i         l.e. float  Weights for obs. i and its neighbors 8*m
     W_SUM_i      l.e. float  Sum of weights for "                 8
    ============ ============ ==================================== ================================

    """  # noqa: E501

    FORMATS = ["swm"]
    MODES = ["r", "w"]

    def __init__(self, *args, **kwargs):
        self._varName = "Unknown"
        self._srs = "Unknown"
        fileio.FileIO.__init__(self, *args, **kwargs)
        self.file = open(self.dataPath, self.mode + "b")

    def _set_varName(self, val):
        if issubclass(type(val), str):
            self._varName = val

    def _get_varName(self):
        return self._varName

    varName = property(fget=_get_varName, fset=_set_varName)

    def _set_srs(self, val):
        if issubclass(type(val), str):
            self._srs = val

    def _get_srs(self):
        return self._srs

    srs = property(fget=_get_srs, fset=_set_srs)

    def read(self, n=-1):  # noqa: ARG002
        self._complain_ifclosed(self.closed)
        return self._read()

    def seek(self, pos):
        if pos == 0:
            self.file.seek(0)
            self.pos = 0

    def _read(self):
        """Read an ArcGIS ``.swm`` file.

        Returns
        -------
        w : libpysal.weights.W
            A PySAL `W` object.

        Raises
        ------
        StopIteration
            Raised at the EOF.

        Examples
        --------

        Type ``dir(w)`` at the interpreter to see what methods are supported.
        Open an ArcGIS ``.swm`` file and read it into a PySAL weights object.

        >>> import libpysal
        >>> w = libpysal.io.open(libpysal.examples.get_path('ohio.swm'), 'r').read()

        Get the number of observations from the header,

        >>> w.n
        88

        Get the mean number of neighbors.

        >>> w.mean_neighbors
        5.25

        Get neighbor distances for a single observation.

        >>> w[1] == dict({2: 1.0, 11: 1.0, 6: 1.0, 7: 1.0})
        True

        """

        if self.pos > 0:
            raise StopIteration

        header = self.file.readline()
        header = header.decode()

        if header.upper().strip().startswith("VERSION@"):
            #  deal with the new SWM version
            w = self.read_new_version(header)
        else:
            #  deal with the old SWM version
            w = self.read_old_version(header)

        return w

    def read_old_version(self, header):
        """Read the old version of ArcGIS(<10.1) ``.swm`` file.

        Parameters
        ----------
        header : str
            The first line of the ``.swm`` file.

        Returns
        -------
        w : libpysal.weights.W
            A PySAL `W` object.

        """

        id_var, srs = header[:-1].split(";")
        self.varName = id_var
        self.srs = srs
        self.header_len = len(header) + 8
        no_obs, row_std = tuple(unpack("<2l", self.file.read(8)))
        neighbors = {}
        weights = {}

        for _ in range(no_obs):
            origin, no_nghs = tuple(unpack("<2l", self.file.read(8)))
            neighbors[origin] = []
            weights[origin] = []

            if no_nghs > 0:
                neighbors[origin] = list(
                    unpack("<%il" % no_nghs, self.file.read(4 * no_nghs))
                )
                weights[origin] = list(
                    unpack("<%id" % no_nghs, self.file.read(8 * no_nghs))
                )
                _ = list(unpack("<d", self.file.read(8)))[0]

        self.pos += 1

        w = W(neighbors, weights)

        return w

    def read_new_version(self, header_line):
        """Read the new version of ArcGIS(<10.1) ``.swm`` file, which contains
        more parameters and records weights in two ways, fixed or variable.

        Parameters
        ----------
        header_line : str
            The first line of the ``.swm`` file, which contains a lot of
            parameters. The parameters are divided by semicolons (';') and
            the key-value of each parameter is divided by at marks ('@').

        Returns
        -------
        w : libpysal.weights.W
            A PySAL `W` object.

        """

        headerDict = {}
        for item in header_line.split(";"):
            key, value = item.split("@")
            headerDict[key] = value

        # for the reader, in order to generate the PySAL Weight class,
        # only a few of the parameters are needed.
        self.varName = headerDict["UNIQUEID"]
        self.srs = headerDict["SPATIALREFNAME"]

        fixedWeights = False
        if "FIXEDWEIGHTS" in headerDict:
            fixedWeights = headerDict["FIXEDWEIGHTS"].upper().strip() == "TRUE"

        no_obs, row_std = tuple(unpack("<2l", self.file.read(8)))

        neighbors = {}
        weights = {}

        for _ in range(no_obs):
            origin, no_nghs = tuple(unpack("<2l", self.file.read(8)))
            neighbors[origin] = []
            weights[origin] = []

            if no_nghs > 0:
                neighbors[origin] = list(
                    unpack("<%il" % no_nghs, self.file.read(4 * no_nghs))
                )

                if fixedWeights:
                    weights[origin] = list(unpack("<d", self.file.read(8))) * no_nghs
                else:
                    weights[origin] = list(
                        unpack("<%id" % no_nghs, self.file.read(8 * no_nghs))
                    )
                _ = list(unpack("<d", self.file.read(8)))[0]

        self.pos += 1

        w = W(neighbors, weights)

        return w

    def write(self, obj, useIdIndex=False):  # noqa: N803
        """Writes a spatial weights matrix data file in ``.swm`` format.

        Parameters
        ----------
        obj : libpysal.weights.W
            A PySAL `W` object.
        useIdIndex : bool
            Use the `W` IDs and remap (``True``). Default is ``False``.

        Raises
        ------
        TypeError
            Raised when the input ``obj`` is not a PySAL `W`.
        TypeError
            Raised when the IDs in input ``obj`` are not integers.

        Examples
        --------

        >>> import tempfile, libpysal, os
        >>> testfile = libpysal.io.open(libpysal.examples.get_path('ohio.swm'), 'r')
        >>> w = testfile.read()

        Create a temporary file for this example.

        >>> f = tempfile.NamedTemporaryFile(suffix='.swm')

        Reassign to a new variable.

        >>> fname = f.name

        Close the temporary named file.

        >>> f.close()

        Open the new file in write mode.

        >>> o = libpysal.io.open(fname, 'w')

        Add properities to the file to write.

        >>> o.varName = testfile.varName
        >>> o.srs = testfile.srs

        Write the weights object into the open file.

        >>> o.write(w)
        >>> o.close()

        Read in the newly created text file.

        >>> wnew = libpysal.io.open(fname, 'r').read()

        Compare values from old to new.

        >>> wnew.pct_nonzero == w.pct_nonzero
        True

        Clean up the temporary file created for this example.

        >>> os.remove(fname)

        """

        self._complain_ifclosed(self.closed)

        if not issubclass(type(obj), W):
            raise TypeError("Expected a PySAL weights object, got: %s." % (type(obj)))

        if (type(obj.id_order[0]) not in (np.int32, np.int64, int)) and not useIdIndex:
            raise TypeError("ArcGIS SWM files support only integer IDs.")

        if useIdIndex:
            id2i = obj.id2i
            obj = remap_ids(obj, id2i)

        unk = str(f"{self.varName};{self.srs}\n").encode()
        self.file.write(unk)
        self.file.write(pack("<l", obj.n))
        self.file.write(pack("<l", obj.transform.upper() == "R"))

        for obs in obj.weights:
            self.file.write(pack("<l", obs))
            no_nghs = len(obj.weights[obs])
            self.file.write(pack("<l", no_nghs))
            self.file.write(pack(f"<{no_nghs}l", *obj.neighbors[obs]))
            self.file.write(pack(f"<{no_nghs}d", *obj.weights[obs]))
            self.file.write(pack("<d", sum(obj.weights[obs])))

        self.pos += 1

    def close(self):
        self.file.close()
        fileio.FileIO.close(self)



""" Load WKB into PySAL shapes.

Where PySAL shapes support multiple parts, "MULTI"type shapes
will be converted to a single multi-part shape:

    MULTIPOLYGON -> Polygon
    MULTILINESTRING -> Chain

Otherwise a list of shapes will be returned:

    MULTIPOINT -> [pt0, ..., ptN]

Some concepts aren't well supported by PySAL shapes. For example:

    wkt = 'MULTIPOLYGON EMPTY' -> '\x01   \x06\x00\x00\x00   \x00\x00\x00\x00'
                                  |  <  | WKBMultiPolygon |    0 parts      |

``pysal.cg.Polygon`` does not support 0 part polygons. ``None`` is returned in this case.


REFERENCE MATERIAL:
SOURCE: http://webhelp.esri.com/arcgisserver/9.3/dotNet/index.htm#geodatabases/the_ogc_103951442.htm

    Basic Type definitions
    byte : 1 byte
    uint32 : 32 bit unsigned integer  (4 bytes)
    double : double precision number (8 bytes)

    Building Blocks : Point, LinearRing

"""  # noqa: E501

import struct
import sys
from io import StringIO

from ... import cg

__author__ = "Charles R Schmidt <schmidtc@gmail.com>"
__all__ = ["loads"]

"""
enum wkbByteOrder {
    wkbXDR = 0,              Big Endian
    wkbNDR = 1               Little Endian
};
"""


DEFAULT_ENDIAN = "<" if sys.byteorder == "little" else ">"
ENDIAN = {"\x00": ">", "\x01": "<"}


def load_ring_little(dat) -> list:
    """
    LinearRing   {
        uint32  numPoints;
        Point   points[numPoints];
    }
    """
    npts = struct.unpack("<I", dat.read(4))[0]
    xy = struct.unpack("<%dd" % (npts * 2), dat.read(npts * 2 * 8))
    return [cg.Point(xy[i : i + 2]) for i in range(0, npts * 2, 2)]


def load_ring_big(dat) -> list:
    npts = struct.unpack(">I", dat.read(4))[0]
    xy = struct.unpack(">%dd" % (npts * 2), dat.read(npts * 2 * 8))
    return [cg.Point(xy[i : i + 2]) for i in range(0, npts * 2, 2)]


def loads(s: str):
    """
    WKBGeometry  {
        union {
            WKBPoint                        point;
            WKBLineString                   linestring;
            WKBPolygon                      polygon;
            WKBGeometryCollection           collection;
            WKBMultiPoint                   mpoint;
            WKBMultiLineString              mlinestring;
            WKBMultiPolygon                 mpolygon;
        }
    };

    Returns
    -------
    geom : {None, libpysal.cg.{Point, Chain, Polygon}}
        The geometric object or ``None``.

    Raises
    ------
    TypeError
        Raised when an unsupported shape type is passed in.

    """

    # To allow recursive calls, read only the bytes we need.
    dat = s if hasattr(s, "read") else StringIO(s)
    endian = ENDIAN[dat.read(1)]
    typ = struct.unpack("I", dat.read(4))[0]
    if typ == 1:
        """
        WKBPoint {
            byte                byteOrder;
            uint32              wkbType;                        1
            Point               point;
        }
        Point {
            double x;
            double y;
        };
        """
        x, y = struct.unpack(endian + "dd", dat.read(16))
        geom = cg.Point((x, y))

    elif typ == 2:
        """
        WKBLineString {
            byte                byteOrder;
            uint32              wkbType;                        2
            uint32              numPoints;
            Point               points[numPoints];
        }
        """
        n = struct.unpack(endian + "I", dat.read(4))[0]
        xy = struct.unpack(endian + "%dd" % (n * 2), dat.read(n * 2 * 8))
        geom = cg.Chain([cg.Point(xy[i : i + 2]) for i in range(0, n * 2, 2)])

    elif typ == 3:
        """
        WKBPolygon  {
            byte                byteOrder;
            uint32              wkbType;                        3
            uint32              numRings;
            LinearRing          rings[numRings];
        }

        WKBPolygon has exactly 1 outer ring and `n` holes.
        Multipart Polygons are NOT support by WKBPolygon.
        """
        nrings = struct.unpack(endian + "I", dat.read(4))[0]
        load_ring = load_ring_little if endian == "<" else load_ring_big
        rings = [load_ring(dat) for _ in range(nrings)]
        geom = cg.Polygon(rings[0], rings[1:])
    elif typ == 4:
        """
        WKBMultiPoint   {
            byte                byteOrder;
            uint32              wkbType;                        4
            uint32              num_wkbPoints;
            WKBPoint            WKBPoints[num_wkbPoints];
        }
        """
        npts = struct.unpack(endian + "I", dat.read(4))[0]
        geom = [loads(dat) for _ in range(npts)]

    elif typ == 5:
        """
        WKBMultiLineString  {
            byte                byteOrder;
            uint32              wkbType;                        5
            uint32              num_wkbLineStrings;
            WKBLineString       WKBLineStrings[num_wkbLineStrings];
        }
        """
        nparts = struct.unpack(endian + "I", dat.read(4))[0]
        chains = [loads(dat) for _ in range(nparts)]
        geom = cg.Chain(sum([c.parts for c in chains], []))

    elif typ == 6:
        """
        wkbMultiPolygon {
            byte                byteOrder;
            uint32              wkbType;                        6
            uint32              num_wkbPolygons;
            WKBPolygon          wkbPolygons[num_wkbPolygons];
        }

        """
        npolys = struct.unpack(endian + "I", dat.read(4))[0]
        polys = [loads(dat) for _ in range(npolys)]
        parts = sum([p.parts for p in polys], [])
        holes = sum([p.holes for p in polys if p.holes[0]], [])

        # MULTIPOLYGON EMPTY, isn't well supported by PySAL shape types.
        geom = None if not parts else cg.Polygon(parts, holes)
    elif typ == 7:
        """
        WKBGeometryCollection {
            byte                byte_order;
            uint32              wkbType;                        7
            uint32              num_wkbGeometries;
            WKBGeometry         wkbGeometries[num_wkbGeometries]
        }
        """
        ngeoms = struct.unpack(endian + "I", dat.read(4))[0]
        geom = [loads(dat) for _ in range(ngeoms)]

    try:
        return geom
    except NameError:
        raise TypeError("Type (%d) is unknown or unsupported." % typ) from None


# if __name__ == "__main__":
#
#    # TODO: Refactor below into Unit Tests
#    wktExamples = [
#        "POINT(6 10)",
#        "LINESTRING(3 4,10 50,20 25)",
#        "POLYGON((1 1,5 1,5 5,1 5,1 1),(2 2, 3 2, 3 3, 2 3,2 2))",
#        "MULTIPOINT(3.5 5.6,4.8 10.5)",
#        "MULTILINESTRING((3 4,10 50,20 25),(-5 -8,-10 -8,-15 -4))",
#        # This MULTIPOLYGON is not valid, the 2nd shell instects the 1st.
#        #(
#        #    'MULTIPOLYGON(((1 1,5 1,5 5,1 5,1 1),"
#        #    "(2 2, 3 2, 3 3, 2 3,2 2)),((3 3,6 2,6 4,3 3)))'
#        #,
#        (
#            "MULTIPOLYGON(((1 1,5 1,5 5,1 5,1 1),"
#            "(2 2, 3 2, 3 3, 2 3,2 2)),((5 3,6 2,6 4,5 3)))"
#        ),
#        "GEOMETRYCOLLECTION(POINT(4 6),LINESTRING(4 6,7 10))",
#        #'POINT ZM (1 1 5 60)',  <-- ZM is not supported by WKB ?
#        #'POINT M (1 1 80)',     <-- M is not supported by WKB ?
#        #'POINT EMPTY',          <-- NOT SUPPORT
#        "MULTIPOLYGON EMPTY",
#    ]
#
#    # shapely only used for testing.
#    try:
#        import shapely.geometry
#        import shapely.wkt
#        from pysal.contrib.shapely_ext import to_wkb
#    except ImportError:
#        print("shapely is used to test this module.")
#        raise
#    for example in wktExamples:
#        print(example)
#        shape0 = shapely.wkt.loads(example)
#        shape1 = loads(shape0.to_wkb())
#        if example.startswith("MULTIPOINT"):
#            shape2 = shapely.geometry.asMultiPoint(shape1)
#        elif example.startswith("GEOMETRYCOLLECTION"):
#            shape2 = shapely.geometry.collection.GeometryCollection(
#                list(map(shapely.geometry.asShape, shape1))
#            )
#        elif example == "MULTIPOLYGON EMPTY":
#            # Skip Test
#            shape2 = None
#        else:
#            shape2 = shapely.geometry.asShape(shape1)
#
#        print(shape1)
#        if shape2:
#            assert shape0.equals(shape2)
#            print(shape0.equals(shape2))
#        else:
#            print("Skip")
#
#        print("")



"""
For completeness, we need to test a shuffled dataframe
(i.e. always send unsorted data) with:
- numeric ids
- string ids
- point dataframe
- coordinates
- check two kernel functions
- check two tree types
- scikit/no scikit
"""

import geodatasets
import geopandas
import numpy as np
import pandas as pd
import pytest

from libpysal.graph._kernel import (
    HAS_SKLEARN,
    _distance_band,
    _kernel,
    _kernel_functions,
)

grocs = geopandas.read_file(geodatasets.get_path("geoda groceries"))[
    ["OBJECTID", "geometry"]
].explode(ignore_index=True)
grocs["strID"] = grocs.OBJECTID.astype(str)
grocs["intID"] = grocs.OBJECTID.values

kernel_functions = list(_kernel_functions.keys())


def my_kernel(distances, bandwidth):
    output = np.cos(distances / distances.max())
    output[distances < bandwidth] = 0
    return output


kernel_functions.append(my_kernel)

metrics = ("euclidean", "haversine")

np.random.seed(6301)
# create a 2-d laplace distribution as a "degenerate"
# over-concentrated distribution
# and rescale to match the lenght-scale in groceries
lap_coords = np.random.laplace(size=(200, 2)) / 50

# create a 2-d cauchy as a "degenerate"
# spatial outlier-y distribution
cau_coords = np.random.standard_cauchy(size=(200, 2))

data = (grocs, lap_coords)

parametrize_ids = pytest.mark.parametrize("ids", [None, "strID", "intID"])
parametrize_data = pytest.mark.parametrize("data", [grocs, lap_coords, cau_coords])
parametrize_kernelfunctions = pytest.mark.parametrize("kernel", kernel_functions)
parametrize_metrics = pytest.mark.parametrize("metric", metrics, metrics)

# how do we parameterize conditional on sklearn in env?


@parametrize_ids
def test_neighbors(ids):
    data = grocs.set_index(ids) if ids else grocs
    head, tail, weight = _kernel(data, bandwidth=5000, kernel="boxcar")
    assert head.shape[0] == 437
    assert tail.shape == head.shape
    assert weight.shape == head.shape
    np.testing.assert_array_equal(pd.unique(head), data.index)
    known = np.linspace(9, 436, 10, dtype=int)
    head_exp = [2, 16, 28, 41, 55, 72, 92, 111, 135, 147]
    if ids:
        head_exp = data.index.values[head_exp]
    np.testing.assert_array_equal(head.values[known], head_exp)
    tail_exp = [13, 92, 66, 31, 33, 73, 16, 103, 9, 147]
    if ids:
        tail_exp = data.index.values[tail_exp]
    np.testing.assert_array_equal(tail.values[known], tail_exp)
    np.testing.assert_array_equal(
        weight[known],
        np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 0], dtype="int64"),
    )


@parametrize_data
def test_no_taper(data):
    head, tail, weight = _kernel(data, taper=False)
    assert head.shape[0] == len(data) * (len(data) - 1)
    assert tail.shape == head.shape
    assert weight.shape == head.shape
    if hasattr(data, "index"):
        np.testing.assert_array_equal(np.unique(head), data.index)
    else:
        np.testing.assert_array_equal(np.unique(head), np.arange(len(data)))


@parametrize_ids
def test_ids(ids):
    data = grocs.set_index(ids) if ids else grocs
    head, tail, _ = _kernel(data)
    np.testing.assert_array_equal(pd.unique(head), data.index)
    assert np.in1d(tail, data.index).all()


def test_distance():
    _, _, weight = _kernel(grocs, kernel="identity")
    known = np.linspace(9, weight.shape[0], 10, dtype=int, endpoint=False)
    np.testing.assert_array_almost_equal(
        weight[known],
        np.array(
            [
                39028.10991144,
                51086.85388002,
                21270.55278224,
                8999.11607504,
                91203.25966722,
                36548.75743352,
                58917.81440314,
                63359.35143896,
                24952.48387721,
                65860.55093353,
            ]
        ),
    )

    _, _, weight = _kernel(lap_coords, kernel="identity")
    known = np.linspace(9, weight.shape[0], 10, dtype=int, endpoint=False)
    np.testing.assert_array_almost_equal(
        weight[known],
        np.array(
            [
                0.0112305,
                0.03158595,
                0.06027445,
                0.01274032,
                0.07559474,
                0.02240698,
                0.07024776,
                0.05554498,
                0.06012029,
                0.03836196,
            ]
        ),
    )

    _, _, weight = _kernel(cau_coords, kernel="identity")
    known = np.linspace(9, weight.shape[0], 10, dtype=int, endpoint=False)
    np.testing.assert_array_almost_equal(
        weight[known],
        np.array(
            [
                5.16946327,
                2.29669606,
                1.60345402,
                1.23831204,
                19.2459334,
                2.99545291,
                2.33316209,
                2.09711302,
                2.12594958,
                6.19347592,
            ]
        ),
    )


@parametrize_data
def test_k(data):
    head, tail, weight = _kernel(data, k=3, kernel="identity")
    assert head.shape[0] == data.shape[0] * 3
    assert tail.shape == head.shape
    assert weight.shape == head.shape

    # TODO: what shall be the behaviour when k is set but bandwidth
    # TODO: is too small so weight is zero, eliminating the link with taper?
    # TODO: now the affected focal has less neighbors than K without warning
    # TODO: test it with the code above and default kernel


@parametrize_kernelfunctions
def test_kernels(kernel):
    _, _, weight = _kernel(grocs, kernel=kernel, taper=False)
    if kernel == "triangular":
        assert weight.mean() == pytest.approx(0.09416822598434019)
        assert weight.max() == pytest.approx(0.9874476868358023)
    elif kernel == "parabolic":
        assert weight.mean() == pytest.approx(0.10312196315841769)
        assert weight.max() == pytest.approx(0.749881829575671)
    elif kernel == "gaussian":
        assert weight.mean() == pytest.approx(0.1124559308071747)
        assert weight.max() == pytest.approx(0.22507021331712948)
    elif kernel == "bisquare":
        assert weight.mean() == pytest.approx(0.09084085210598618)
        assert weight.max() == pytest.approx(0.9372045972129259)
    elif kernel == "cosine":
        assert weight.mean() == pytest.approx(0.1008306468068958)
        assert weight.max() == pytest.approx(0.7852455006403666)
    elif kernel in ["boxcar", "discrete"]:
        assert weight.mean() == pytest.approx(0.2499540356683214)
        assert weight.max() == 1
    elif kernel in ["identity", None]:
        assert weight.mean() == pytest.approx(39758.007361814016)
        assert weight.max() == pytest.approx(127937.75271993055)
    else:  # function
        assert weight.mean() == pytest.approx(0.6880384553732511)
        assert weight.max() == pytest.approx(0.9855481738848647)


@parametrize_data
@pytest.mark.parametrize("bandwidth", [None, 0.05, 0.4])
def test_bandwidth(data, bandwidth):
    head, tail, weight = _kernel(data, bandwidth=bandwidth)
    assert tail.shape == head.shape
    assert weight.shape == head.shape
    if hasattr(data, "index"):
        np.testing.assert_array_equal(np.unique(head), data.index)
    else:
        np.testing.assert_array_equal(np.unique(head), np.arange(len(data)))


@pytest.mark.parametrize(
    "metric",
    [
        "euclidean",
        "minkowski",
        "cityblock",
        "chebyshev",
        "haversine",
    ],
)
def test_metric(metric):
    data = grocs.to_crs(4326) if metric == "haversine" else grocs
    if not HAS_SKLEARN and metric in ["chebyshev", "haversine"]:
        pytest.skip("metric not supported by scipy")
    head, tail, weight = _kernel(data, metric=metric, kernel="identity", p=1.5)
    assert head.shape[0] == len(data) * (len(data) - 1)
    assert tail.shape == head.shape
    assert weight.shape == head.shape
    np.testing.assert_array_equal(pd.unique(head), data.index)

    if metric == "euclidean":
        assert weight.mean() == pytest.approx(39758.007362)
        assert weight.max() == pytest.approx(127937.75272)
    elif metric == "minkowski":
        assert weight.mean() == pytest.approx(42288.642129)
        assert weight.max() == pytest.approx(140674.095752)
    elif metric == "cityblock":
        assert weight.mean() == pytest.approx(49424.576155)
        assert weight.max() == pytest.approx(173379.431622)
    elif metric == "chebyshev":
        assert weight.mean() == pytest.approx(36590.352895)
        assert weight.max() == pytest.approx(123955.14249)
    else:
        assert weight.mean() == pytest.approx(0.115835)
        assert weight.max() == pytest.approx(0.371465)


@pytest.mark.parametrize(
    "metric",
    [
        "euclidean",
        "minkowski",
        "cityblock",
        "chebyshev",
        "haversine",
    ],
)
def test_metric_k(metric):
    data = grocs.to_crs(4326) if metric == "haversine" else grocs
    if not HAS_SKLEARN and metric in ["chebyshev", "haversine"]:
        pytest.skip("metric not supported by scipy")
    head, tail, weight = _kernel(data, k=3, metric=metric, kernel="identity", p=1.5)
    assert head.shape[0] == len(data) * 3
    assert tail.shape == head.shape
    assert weight.shape == head.shape
    np.testing.assert_array_equal(pd.unique(head), data.index)

    if metric == "euclidean":
        assert weight.mean() == pytest.approx(4577.237441)
        assert weight.max() == pytest.approx(18791.085051)
    elif metric == "minkowski":
        assert weight.mean() == pytest.approx(4884.254721)
        assert weight.max() == pytest.approx(20681.125211)
    elif metric == "cityblock":
        assert weight.mean() == pytest.approx(5665.288523)
        assert weight.max() == pytest.approx(23980.893147)
    elif metric == "chebyshev":
        assert weight.mean() == pytest.approx(4032.283559)
        assert weight.max() == pytest.approx(16374.141739)
    else:
        assert weight.mean() == pytest.approx(0.00021882448)
        assert weight.max() == pytest.approx(0.000897441)


# def test_precomputed(data, ids):
#     raise NotImplementedError()


def test_coincident():
    grocs_duplicated = pd.concat(
        [grocs, grocs.iloc[:10], grocs.iloc[:3]], ignore_index=True
    )
    # plain kernel
    head, tail, weight = _kernel(grocs_duplicated)
    assert head.shape[0] == len(grocs_duplicated) * (len(grocs_duplicated) - 1)
    assert tail.shape == head.shape
    assert weight.shape == head.shape
    np.testing.assert_array_equal(pd.unique(head), grocs_duplicated.index)

    # k, raise
    with pytest.raises(ValueError, match="There are"):
        _kernel(grocs_duplicated, k=2)

    # k, jitter
    head, tail, weight = _kernel(
        grocs_duplicated, taper=False, k=2, coincident="jitter"
    )
    assert head.shape[0] == len(grocs_duplicated) * 2
    assert tail.shape == head.shape
    assert weight.shape == head.shape
    np.testing.assert_array_equal(pd.unique(head), grocs_duplicated.index)

    # k, clique
    head, tail, weight = _kernel(grocs_duplicated, k=2, coincident="clique")
    assert head.shape[0] >= len(grocs_duplicated) * 2
    assert tail.shape == head.shape
    assert weight.shape == head.shape
    np.testing.assert_array_equal(pd.unique(head), grocs_duplicated.index)


def test_shape_preservation():
    coordinates = np.vstack(
        [np.repeat(np.arange(10), 10), np.tile(np.arange(10), 10)]
    ).T
    head, tail, weight = _kernel(
        coordinates,
        k=3,
        metric="euclidean",
        p=2,
        kernel="boxcar",
        bandwidth=0.5,
        taper=False,
    )
    np.testing.assert_array_equal(head, np.repeat(np.arange(100), 3))
    assert tail.shape == head.shape, "shapes of head and tail do not match"
    np.testing.assert_array_equal(weight, np.zeros((300,), dtype=int))

    head, tail, weight = _kernel(
        coordinates,
        k=3,
        metric="euclidean",
        p=2,
        kernel="boxcar",
        bandwidth=0.5,
        taper=True,
    )
    np.testing.assert_array_equal(head, np.arange(100))
    assert tail.shape == head.shape, "shapes of head and tail do not match"
    np.testing.assert_array_equal(weight, np.zeros((100,), dtype=int))


def test_haversine_check():
    with pytest.raises(ValueError, match="'haversine'"):
        _kernel(grocs, k=2, metric="haversine")


def test_distance_band_colocated():
    coordinates = np.array([[0, 0], [1, 0], [1, 0], [2, 0], [3, 0]])
    dist = _distance_band(coordinates, 1)
    assert dist.shape == (5, 5)
    np.testing.assert_array_equal(
        dist.data,
        np.array(
            [
                0.0,
                1.0,
                1.0,
                1.0,
                0.0,
                0.0,
                1.0,
                1.0,
                0.0,
                0.0,
                1.0,
                1.0,
                1.0,
                0.0,
                1.0,
                1.0,
                0.0,
            ]
        ),
    )



import pytest

from ..shapes import Chain, Line, LineSegment, Point, Polygon, Ray, Rectangle


class TesttestPoint:
    def test___init__1(self):
        """Tests whether points are created without issue."""

        for l_ in [(-5.0, 10.0), (0.0, -6.0), (1e300, float(-1e300))]:
            Point(l_)

    def test___str__1(self):
        """Tests whether the string produced is valid for corner cases."""

        for l_ in [(-5, 10), (0, -6.0), (1e300, -1e300)]:
            p = Point(l_)
            # Recast to floats like point does
            assert str(p) == str((float(l_[0]), float(l_[1])))


class TesttestLineSegment:
    def test_is_ccw1(self):
        """Test corner cases for horizontal segment starting at origin."""

        ls = LineSegment(Point((0, 0)), Point((5, 0)))

        # At positive boundary beyond segment
        assert not ls.is_ccw(Point((10, 0)))
        # On segment
        assert not ls.is_ccw(Point((3, 0)))
        # At negative boundary beyond segment
        assert not ls.is_ccw(Point((-10, 0)))
        # Endpoint of segment
        assert not ls.is_ccw(Point((0, 0)))
        # Endpoint of segment
        assert not ls.is_ccw(Point((5, 0)))

    def test_is_ccw2(self):
        """Test corner cases for vertical segment ending at origin."""

        ls = LineSegment(Point((0, -5)), Point((0, 0)))

        # At positive boundary beyond segment
        assert not ls.is_ccw(Point((0, 10)))
        # On segment
        assert not ls.is_ccw(Point((0, -3)))
        # At negative boundary beyond segment
        assert not ls.is_ccw(Point((0, -10)))
        # Endpoint of segment
        assert not ls.is_ccw(Point((0, -5)))
        # Endpoint of segment
        assert not ls.is_ccw(Point((0, 0)))

    def test_is_ccw3(self):
        """Test corner cases for non-axis-aligned segment not through origin."""

        ls = LineSegment(Point((0, 1)), Point((5, 6)))

        # At positive boundary beyond segment
        assert not ls.is_ccw(Point((10, 11)))
        # On segment
        assert not ls.is_ccw(Point((3, 4)))
        # At negative boundary beyond segment
        assert not ls.is_ccw(Point((-10, -9)))
        # Endpoint of segment
        assert not ls.is_ccw(Point((0, 1)))
        # Endpoint of segment
        assert not ls.is_ccw(Point((5, 6)))

    def test_is_cw1(self):
        """Test corner cases for horizontal segment starting at origin."""

        ls = LineSegment(Point((0, 0)), Point((5, 0)))

        # At positive boundary beyond segment
        assert not ls.is_cw(Point((10, 0)))
        # On segment
        assert not ls.is_cw(Point((3, 0)))
        # At negative boundary beyond segment
        assert not ls.is_cw(Point((-10, 0)))
        # Endpoint of segment
        assert not ls.is_cw(Point((0, 0)))
        # Endpoint of segment
        assert not ls.is_cw(Point((5, 0)))

    def test_is_cw2(self):
        """Test corner cases for vertical segment ending at origin."""

        ls = LineSegment(Point((0, -5)), Point((0, 0)))

        # At positive boundary beyond segment
        assert not ls.is_cw(Point((0, 10)))
        # On segment
        assert not ls.is_cw(Point((0, -3)))
        # At negative boundary beyond segment
        assert not ls.is_cw(Point((0, -10)))
        # Endpoint of segment
        assert not ls.is_cw(Point((0, -5)))
        # Endpoint of segment
        assert not ls.is_cw(Point((0, 0)))

    def test_is_cw3(self):
        """Test corner cases for non-axis-aligned segment not through origin."""

        ls = LineSegment(Point((0, 1)), Point((5, 6)))

        # At positive boundary beyond segment
        assert not ls.is_cw(Point((10, 11)))
        # On segment
        assert not ls.is_cw(Point((3, 4)))
        # At negative boundary beyond segment
        assert not ls.is_cw(Point((-10, -9)))
        # Endpoint of segment
        assert not ls.is_cw(Point((0, 1)))
        # Endpoint of segment
        assert not ls.is_cw(Point((5, 6)))

    def test_get_swap1(self):
        """Tests corner cases."""

        ls = LineSegment(Point((0, 0)), Point((10, 0)))
        swap = ls.get_swap()
        assert ls.p1 == swap.p2
        assert ls.p2 == swap.p1

        ls = LineSegment(Point((-5, 0)), Point((5, 0)))
        swap = ls.get_swap()
        assert ls.p1 == swap.p2
        assert ls.p2 == swap.p1

        ls = LineSegment(Point((0, 0)), Point((0, 0)))
        swap = ls.get_swap()
        assert ls.p1 == swap.p2
        assert ls.p2 == swap.p1

        ls = LineSegment(Point((5, 5)), Point((5, 5)))
        swap = ls.get_swap()
        assert ls.p1 == swap.p2
        assert ls.p2 == swap.p1

    def test_bounding_box(self):
        """Tests corner cases."""

        ls = LineSegment(Point((0, 0)), Point((0, 10)))
        assert ls.bounding_box.left == 0
        assert ls.bounding_box.lower == 0
        assert ls.bounding_box.right == 0
        assert ls.bounding_box.upper == 10

        ls = LineSegment(Point((0, 0)), Point((-3, -4)))
        assert ls.bounding_box.left == -3
        assert ls.bounding_box.lower == -4
        assert ls.bounding_box.right == 0
        assert ls.bounding_box.upper == 0

        ls = LineSegment(Point((-5, 0)), Point((3, 0)))
        assert ls.bounding_box.left == -5
        assert ls.bounding_box.lower == 0
        assert ls.bounding_box.right == 3
        assert ls.bounding_box.upper == 0

    def test_len1(self):
        """Tests corner cases."""

        ls = LineSegment(Point((0, 0)), Point((0, 0)))
        assert ls.len == 0

        ls = LineSegment(Point((0, 0)), Point((-3, 0)))
        assert ls.len == 3

    def test_line1(self):
        """Tests corner cases."""

        import math

        ls = LineSegment(Point((0, 0)), Point((1, 0)))
        assert ls.line.m == 0
        assert ls.line.b == 0

        ls = LineSegment(Point((0, 0)), Point((0, 1)))
        assert ls.line.m == float("inf")
        assert math.isnan(ls.line.b)

        ls = LineSegment(Point((0, 0)), Point((0, -1)))
        assert ls.line.m == float("inf")
        assert math.isnan(ls.line.b)

        ls = LineSegment(Point((0, 0)), Point((0, 0)))
        assert ls.line is None

        ls = LineSegment(Point((5, 0)), Point((10, 0)))
        ls1 = LineSegment(Point((5, 0)), Point((10, 1)))
        assert ls.intersect(ls1)
        ls2 = LineSegment(Point((5, 1)), Point((10, 1)))
        assert not ls.intersect(ls2)
        ls2 = LineSegment(Point((7, -1)), Point((7, 2)))
        assert ls.intersect(ls2)


class TesttestLine:
    def test___init__1(self):
        """Tests a variety of generic cases."""

        for m, b in [(4, 0.0), (-140, 5), (0, 0)]:
            _ = Line(m, b)

    def test_y1(self):
        """Tests a variety of generic and special cases (+-infinity)."""

        l_ = Line(0, 0)
        assert l_.y(0) == 0
        assert l_.y(-1e600) == 0
        assert l_.y(1e600) == 0

        l_ = Line(1, 1)
        assert l_.y(2) == 3
        assert l_.y(-1e600) == -1e600
        assert l_.y(1e600) == 1e600

        l_ = Line(-1, 1)
        assert l_.y(2) == -1
        assert l_.y(-1e600) == 1e600
        assert l_.y(1e600) == -1e600

    def test_x1(self):
        """Tests a variety of generic and special cases (+-infinity)."""

        l_ = Line(0, 0)

        # self.assertEquals(l.x(0), 0)
        with pytest.raises(ArithmeticError):
            l_.x(0)
        with pytest.raises(ArithmeticError):
            l_.x(-1e600)
        with pytest.raises(ArithmeticError):
            l_.x(1e600)

        l_ = Line(1, 1)
        assert l_.x(3) == 2
        assert l_.x(-1e600) == -1e600
        assert l_.x(1e600) == 1e600

        l_ = Line(-1, 1)
        assert l_.x(2) == -1
        assert l_.x(-1e600) == 1e600
        assert l_.x(1e600) == -1e600


class TesttestRay:
    def test___init__1(self):
        """Tests generic cases."""

        _ = Ray(Point((0, 0)), Point((1, 1)))
        _ = Ray(Point((8, -3)), Point((-5, 9)))


class TesttestChain:
    def test___init__1(self):
        """Generic testing that no exception is thrown."""

        _ = Chain([Point((0, 0))])
        _ = Chain([[Point((0, 0)), Point((1, 1))], [Point((2, 5))]])

    def test_vertices1(self):
        """Testing for repeated vertices and multiple parts."""

        vertices = [
            Point((0, 0)),
            Point((1, 1)),
            Point((2, 5)),
            Point((0, 0)),
            Point((1, 1)),
            Point((2, 5)),
        ]
        assert Chain(vertices).vertices == vertices

        vertices = [
            [Point((0, 0)), Point((1, 1)), Point((2, 5))],
            [Point((0, 0)), Point((1, 1)), Point((2, 5))],
        ]
        assert Chain(vertices).vertices == vertices[0] + vertices[1]

    def test_parts1(self):
        """Generic testing of parts functionality."""

        vertices = [
            Point((0, 0)),
            Point((1, 1)),
            Point((2, 5)),
            Point((0, 0)),
            Point((1, 1)),
            Point((2, 5)),
        ]
        assert Chain(vertices).parts == [vertices]

        vertices = [
            [Point((0, 0)), Point((1, 1)), Point((2, 5))],
            [Point((0, 0)), Point((1, 1)), Point((2, 5))],
        ]
        assert Chain(vertices).parts == vertices

    def test_bounding_box1(self):
        """Test correctness with multiple parts."""

        vertices = [
            [Point((0, 0)), Point((1, 1)), Point((2, 6))],
            [Point((-5, -5)), Point((0, 0)), Point((2, 5))],
        ]
        bb = Chain(vertices).bounding_box
        assert bb.left == -5
        assert bb.lower == -5
        assert bb.right == 2
        assert bb.upper == 6

    def test_len1(self):
        """Test correctness with multiple parts and
        zero-length point-to-point distances.

        """

        vertices = [
            [Point((0, 0)), Point((1, 0)), Point((1, 5))],
            [Point((-5, -5)), Point((-5, 0)), Point((0, 0)), Point((0, 0))],
        ]
        assert Chain(vertices).len == 6 + 10


class TesttestPolygon:
    def test___init__1(self):
        """Test various input configurations (list vs. lists of lists, holes)."""

        # Input configurations tested (in order of test):
        # one part, no holes
        # multi parts, no holes
        # one part, one hole
        # multi part, one hole
        # one part, multi holes
        # multi part, multi holes

        _ = Polygon([Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))])
        _ = Polygon(
            [
                [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
                [Point((30, 30)), Point((40, 30)), Point((40, 40)), Point((30, 40))],
            ]
        )
        _ = Polygon(
            [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
            holes=[Point((2, 2)), Point((4, 2)), Point((4, 4)), Point((2, 4))],
        )
        _ = Polygon(
            [
                [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
                [Point((30, 30)), Point((40, 30)), Point((40, 40)), Point((30, 40))],
            ],
            holes=[Point((2, 2)), Point((4, 2)), Point((4, 4)), Point((2, 4))],
        )
        _ = Polygon(
            [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
            holes=[
                [Point((2, 2)), Point((4, 2)), Point((4, 4)), Point((2, 4))],
                [Point((6, 6)), Point((6, 8)), Point((8, 8)), Point((8, 6))],
            ],
        )
        _ = Polygon(
            [
                [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
                [Point((30, 30)), Point((40, 30)), Point((40, 40)), Point((30, 40))],
            ],
            holes=[
                [Point((2, 2)), Point((4, 2)), Point((4, 4)), Point((2, 4))],
                [Point((6, 6)), Point((6, 8)), Point((8, 8)), Point((8, 6))],
            ],
        )

    def test_area1(self):
        """Test multiple parts."""

        p = Polygon(
            [
                [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
                [Point((30, 30)), Point((40, 30)), Point((40, 40)), Point((30, 40))],
            ]
        )
        assert p.area == 200

    def test_area2(self):
        """Test holes."""

        p = Polygon(
            [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
            holes=[Point((2, 2)), Point((4, 2)), Point((4, 4)), Point((2, 4))],
        )
        assert p.area == 100 - 4

        p = Polygon(
            [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
            holes=[
                [Point((2, 2)), Point((4, 2)), Point((4, 4)), Point((2, 4))],
                [Point((6, 6)), Point((6, 8)), Point((8, 8)), Point((8, 6))],
            ],
        )
        assert p.area == 100 - (4 + 4)

        p = Polygon(
            [
                [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
                [Point((30, 30)), Point((40, 30)), Point((40, 40)), Point((30, 40))],
            ],
            holes=[
                [Point((2, 2)), Point((4, 2)), Point((4, 4)), Point((2, 4))],
                [Point((36, 36)), Point((36, 38)), Point((38, 38)), Point((38, 36))],
            ],
        )
        assert p.area == 200 - (4 + 4)

    def test_area4(self):
        """Test polygons with vertices in both orders (cw, ccw)."""

        p = Polygon([Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))])
        assert p.area == 100

        p = Polygon([Point((0, 0)), Point((0, 10)), Point((10, 10)), Point((10, 0))])
        assert p.area == 100

    def test_bounding_box1(self):
        """Test polygons with multiple parts."""

        p = Polygon(
            [
                [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
                [Point((30, 30)), Point((40, 30)), Point((40, 40)), Point((30, 40))],
            ]
        )
        bb = p.bounding_box
        assert bb.left == 0
        assert bb.lower == 0
        assert bb.right == 40
        assert bb.upper == 40

    def test_centroid1(self):
        """Test polygons with multiple parts of the same size."""

        p = Polygon(
            [
                [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
                [Point((30, 30)), Point((40, 30)), Point((40, 40)), Point((30, 40))],
            ]
        )
        c = p.centroid
        assert c[0] == 20
        assert c[1] == 20

    def test_centroid2(self):
        """Test polygons with multiple parts of different size."""

        p = Polygon(
            [
                [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
                [Point((30, 30)), Point((35, 30)), Point((35, 35)), Point((30, 35))],
            ]
        )
        c = p.centroid
        assert c[0] == 10.5
        assert c[1] == 10.5

    def test_holes1(self):
        """Test for correct vertex values/order."""

        p = Polygon(
            [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
            holes=[Point((2, 2)), Point((4, 2)), Point((4, 4)), Point((2, 4))],
        )
        assert len(p.holes) == 1
        e_holes = [Point((2, 2)), Point((2, 4)), Point((4, 4)), Point((4, 2))]
        assert p.holes[0] in [
            e_holes,
            [e_holes[-1]] + e_holes[:3],
            e_holes[-2:] + e_holes[:2],
            e_holes[-3:] + [e_holes[0]],
        ]

    def test_holes2(self):
        """Test for multiple holes."""

        p = Polygon(
            [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
            holes=[
                [Point((2, 2)), Point((4, 2)), Point((4, 4)), Point((2, 4))],
                [Point((6, 6)), Point((6, 8)), Point((8, 8)), Point((8, 6))],
            ],
        )
        holes = p.holes
        assert len(holes) == 2

    def test_parts1(self):
        """Test for correct vertex values/order."""

        p = Polygon(
            [
                [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
                [Point((30, 30)), Point((40, 30)), Point((30, 40))],
            ]
        )
        assert len(p.parts) == 2

        part1 = [Point((0, 0)), Point((0, 10)), Point((10, 10)), Point((10, 0))]
        part2 = [Point((30, 30)), Point((30, 40)), Point((40, 30))]
        if len(p.parts[0]) == 4:
            assert p.parts[0] in [
                part1,
                part1[-1:] + part1[:3],
                part1[-2:] + part1[:2],
                part1[-3:] + part1[:1],
            ]
            assert p.parts[1] in [part2, part2[-1:] + part2[:2], part2[-2:] + part2[:1]]
        elif len(p.parts[0]) == 3:
            assert p.parts[0] in [part2, part2[-1:] + part2[:2], part2[-2:] + part2[:1]]
            assert p.parts[1] in [
                part1,
                part1[-1:] + part1[:3],
                part1[-2:] + part1[:2],
                part1[-3:] + part1[:1],
            ]
        else:
            pytest.fail()

    def test_perimeter1(self):
        """Test with multiple parts."""

        p = Polygon(
            [
                [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
                [Point((30, 30)), Point((40, 30)), Point((40, 40)), Point((30, 40))],
            ]
        )
        assert p.perimeter == 80

    def test_perimeter2(self):
        """Test with holes."""

        p = Polygon(
            [
                [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
                [Point((30, 30)), Point((40, 30)), Point((40, 40)), Point((30, 40))],
            ],
            holes=[
                [Point((2, 2)), Point((4, 2)), Point((4, 4)), Point((2, 4))],
                [Point((6, 6)), Point((6, 8)), Point((8, 8)), Point((8, 6))],
            ],
        )
        assert p.perimeter == 80 + 16

    def test_vertices1(self):
        """Test for correct values/order of vertices."""

        p = Polygon([Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))])
        assert len(p.vertices) == 4
        e_verts = [Point((0, 0)), Point((0, 10)), Point((10, 10)), Point((10, 0))]
        assert p.vertices in [
            e_verts,
            e_verts[-1:] + e_verts[:3],
            e_verts[-2:] + e_verts[:2],
            e_verts[-3:] + e_verts[:1],
        ]

    def test_vertices2(self):
        """Test for multiple parts."""

        p = Polygon(
            [
                [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
                [Point((30, 30)), Point((40, 30)), Point((40, 40)), Point((30, 40))],
            ]
        )
        assert len(p.vertices) == 8

    def test_contains_point(self):
        p = Polygon(
            [Point((0, 0)), Point((10, 0)), Point((10, 10)), Point((0, 10))],
            [Point((1, 2)), Point((2, 2)), Point((2, 1)), Point((1, 1))],
        )
        assert p.contains_point((0, 0)) == 0
        assert p.contains_point((1, 1)) == 0
        assert p.contains_point((5, 5)) == 1
        assert p.contains_point((10, 10)) == 0


class TesttestRectangle:
    def test___init__1(self):
        """Test exceptions are thrown correctly."""

        try:
            # right < left
            _ = Rectangle(1, 1, -1, 5)
        except ArithmeticError:
            pass
        else:
            pytest.fail()

        try:
            # upper < lower
            _ = Rectangle(1, 1, 5, -1)
        except ArithmeticError:
            pass
        else:
            pytest.fail()

    def test_set_centroid1(self):
        """Test with rectangles of zero width or height."""

        # Zero width
        r = Rectangle(5, 5, 5, 10)
        r.set_centroid(Point((0, 0)))
        assert r.left == 0
        assert r.lower == -2.5
        assert r.right == 0
        assert r.upper == 2.5

        # Zero height
        r = Rectangle(10, 5, 20, 5)
        r.set_centroid(Point((40, 40)))
        assert r.left == 35
        assert r.lower == 40
        assert r.right == 45
        assert r.upper == 40

        # Zero width and height
        r = Rectangle(0, 0, 0, 0)
        r.set_centroid(Point((-4, -4)))
        assert r.left == -4
        assert r.lower == -4
        assert r.right == -4
        assert r.upper == -4

    def test_set_scale1(self):
        """Test repeated scaling."""

        r = Rectangle(2, 2, 4, 4)

        r.set_scale(0.5)
        assert r.left == 2.5
        assert r.lower == 2.5
        assert r.right == 3.5
        assert r.upper == 3.5

        r.set_scale(2)
        assert r.left == 2
        assert r.lower == 2
        assert r.right == 4
        assert r.upper == 4

    def test_set_scale2(self):
        """Test scaling of rectangles with zero width/height."""

        # Zero width
        r = Rectangle(5, 5, 5, 10)
        r.set_scale(2)
        assert r.left == 5
        assert r.lower == 2.5
        assert r.right == 5
        assert r.upper == 12.5

        # Zero height
        r = Rectangle(10, 5, 20, 5)
        r.set_scale(2)
        assert r.left == 5
        assert r.lower == 5
        assert r.right == 25
        assert r.upper == 5

        # Zero width and height
        r = Rectangle(0, 0, 0, 0)
        r.set_scale(100)
        assert r.left == 0
        assert r.lower == 0
        assert r.right == 0
        assert r.upper == 0

        # Zero width and height
        r = Rectangle(0, 0, 0, 0)
        r.set_scale(0.01)
        assert r.left == 0
        assert r.lower == 0
        assert r.right == 0
        assert r.upper == 0

    def test_area1(self):
        """Test rectangles with zero width/height."""

        # Zero width
        r = Rectangle(5, 5, 5, 10)
        assert r.area == 0

        # Zero height
        r = Rectangle(10, 5, 20, 5)
        assert r.area == 0

        # Zero width and height
        r = Rectangle(0, 0, 0, 0)
        assert r.area == 0

    def test_height1(self):
        """Test rectangles with zero height."""

        # Zero height
        r = Rectangle(10, 5, 20, 5)
        assert r.height == 0

    def test_width1(self):
        """Test rectangles with zero width."""

        # Zero width
        r = Rectangle(5, 5, 5, 10)
        assert r.width == 0



"""
Voronoi tesslation of 2-d point sets.

Adapted from https://gist.github.com/pv/8036995

"""
import warnings

import geopandas as gpd
import numpy as np
import numpy.typing as npt
import shapely
from packaging.version import Version
from scipy.spatial import Voronoi

__author__ = "Serge Rey <sjsrey@gmail.com>"

__all__ = ["voronoi_frames"]

GPD_GE_013 = Version(gpd.__version__) >= Version("0.13.0")


def voronoi(points, radius=None):
    """Determine finite Voronoi diagram for a 2-d point set.
    See also ``voronoi_regions()``.

    Parameters
    ----------
    points : array_like
        An nx2 array of points.
    radius : float (optional)
        The distance to 'points at infinity'. Default is ``None.``

    Returns
    -------
    vor : tuple
        A two-element tuple consisting of a list and an array. Each element of
        the list contains the sequence of the indices of Voronoi vertices
        composing a Voronoi polygon (region), whereas the array contains
        the Voronoi vertex coordinates.

    Examples
    --------
    >>> points = [(10.2, 5.1), (4.7, 2.2), (5.3, 5.7), (2.7, 5.3)]
    >>> regions, coordinates = voronoi(points)
    >>> regions
    [[1, 3, 2], [4, 5, 1, 0], [0, 1, 7, 6], [9, 0, 8]]

    >>> coordinates
    array([[  4.21783296,   4.08408578],
           [  7.51956025,   3.51807539],
           [  9.4642193 ,  19.3994576 ],
           [ 14.98210684, -10.63503022],
           [ -9.22691341,  -4.58994414],
           [ 14.98210684, -10.63503022],
           [  1.78491801,  19.89803294],
           [  9.4642193 ,  19.3994576 ],
           [  1.78491801,  19.89803294],
           [ -9.22691341,  -4.58994414]])
    """
    warnings.warn(
        "The 'voronoi' function is considered private and will be "
        "removed in a future release.",
        FutureWarning,
        stacklevel=2,
    )

    vor = voronoi_regions(Voronoi(points), radius=radius)

    return vor


def voronoi_regions(vor, radius=None):
    """Finite voronoi regions for a 2-d point set. See also ``voronoi()``.

    Parameters
    ----------
    vor : scipy.spatial.Voronoi
        A planar Voronoi diagram.
    radius : float (optional)
        Distance to 'points at infinity'. Default is ``None.``

    Returns
    -------
    regions_vertices : tuple
        A two-element tuple consisting of a list of finite voronoi regions
        and an array Voronoi vertex coordinates.
    """
    warnings.warn(
        "The 'voronoi_regions' function is considered private and will be "
        "removed in a future release.",
        FutureWarning,
        stacklevel=2,
    )

    new_regions = []
    new_vertices = vor.vertices.tolist()

    center = vor.points.mean(axis=0)
    if radius is None:
        radius = vor.points.ptp().max() * 2

    all_ridges = {}
    for (p1, p2), (v1, v2) in zip(vor.ridge_points, vor.ridge_vertices, strict=True):
        all_ridges.setdefault(p1, []).append((p2, v1, v2))
        all_ridges.setdefault(p2, []).append((p1, v1, v2))

    for p1, region in enumerate(vor.point_region):
        vertices = vor.regions[region]

        if all(v >= 0 for v in vertices):
            new_regions.append(vertices)
            continue

        ridges = all_ridges[p1]
        new_region = [v for v in vertices if v >= 0]

        for p2, v1, v2 in ridges:
            if v2 < 0:
                v1, v2 = v2, v1
            if v1 >= 0:
                continue

            t = vor.points[p2] - vor.points[p1]
            t /= np.linalg.norm(t)
            n = np.array([-t[1], t[0]])

            midpoint = vor.points[[p1, p2]].mean(axis=0)
            direction = np.sign(np.dot(midpoint - center, n)) * n
            far_point = vor.vertices[v2] + direction * radius

            new_region.append(len(new_vertices))
            new_vertices.append(far_point.tolist())

        vs = np.asarray([new_vertices[v] for v in new_region])
        c = vs.mean(axis=0)
        angles = np.arctan2(vs[:, 1] - c[1], vs[:, 0] - c[0])
        new_region = np.array(new_region)[np.argsort(angles)]

        new_regions.append(new_region.tolist())

    regions_vertices = new_regions, np.asarray(new_vertices)

    return regions_vertices


def as_dataframes(regions, vertices, points):
    """Helper function to store finite Voronoi regions and
    originator points as ``geopandas`` (or ``pandas``) dataframes.

    Parameters
    ----------
    regions : list
        Each element of the list contains sequence of the indexes of
        voronoi vertices composing a vornoi polygon (region).
    vertices : array_like
        The coordinates of the vornoi vertices.
    points : array_like
        The originator points.

    Returns
    -------
    region_df : geopandas.GeoDataFrame
        Finite Voronoi polygons as geometries.
    points_df : geopandas.GeoDataFrame
        Originator points as geometries.

    Raises
    ------
    ImportError
        Raised when ``geopandas`` is not available.
    ImportError
        Raised when ``shapely`` is not available.
    """
    warnings.warn(
        "The 'as_dataframes' function is considered private and will be "
        "removed in a future release.",
        FutureWarning,
        stacklevel=2,
    )

    try:
        import geopandas as gpd
    except ImportError:
        gpd = None

    try:
        from shapely.geometry import Point, Polygon
    except ImportError:
        from .shapes import Point, Polygon

    if gpd is not None:
        region_df = gpd.GeoDataFrame(
            geometry=gpd.GeoSeries(Polygon(vertices[region]) for region in regions)
        )
        point_df = gpd.GeoDataFrame(
            geometry=gpd.GeoSeries(Point(pnt) for pnt in points)
        )
    else:
        import pandas as pd

        region_df = pd.DataFrame()
        region_df["geometry"] = [
            Polygon(vertices[region].tolist()) for region in regions
        ]
        point_df = pd.DataFrame()
        point_df["geometry"] = [Point(pnt) for pnt in points]

    return region_df, point_df


def clip_voronoi_frames_to_extent(regions, vertices, clip="extent"):
    """Generate a geopandas.GeoDataFrame of Voronoi cells clipped to
    a specified extent.

    Parameters
    ----------
    regions : geopandas.GeoDataFrame
        A (geo)dataframe containing voronoi cells to clip.
    vertices : geopandas.GeoDataFrame
        A (geo)dataframe containing vertices used to build voronoi cells.
    clip : str, shapely.geometry.Polygon
        An overloaded option about how to clip the voronoi cells.
        The options are:
          - 'none'/None: No clip is applied. Voronoi cells may be arbitrarily
            larger that the source map. Note that this may lead to cells that
            are many orders of magnitude larger in extent than
            the original map. Not recommended.
          - 'bbox'/'extent'/'bounding box': Clip the voronoi cells to the
            bounding box of the input points.
          - 'chull'/'convex hull': Clip the voronoi cells to the
            convex hull of the input points.
          - 'ashape'/'ahull': Clip the voronoi cells to the tightest hull that
            contains all points (e.g. the smallest alphashape,
            using ``libpysal.cg.alpha_shape_auto``).
          - Polygon: Clip to an arbitrary Polygon.

    Returns
    -------
    clipped_regions : geopandas.GeoDataFrame
        A ``geopandas.GeoDataFrame`` of clipped voronoi regions.

    Raises
    ------
    ImportError
        Raised when ``shapely`` is not available.
    ImportError
        Raised when ``geopandas`` is not available.
    ValueError
        Raised when in invalid value for ``clip`` is passed in.
    """
    warnings.warn(
        "The 'clip_voronoi_frames_to_extent' function is considered private "
        "and will be removed in a future release.",
        FutureWarning,
        stacklevel=2,
    )

    try:
        from shapely.geometry import Polygon
    except ImportError:
        raise ImportError("Shapely is required to clip voronoi regions.") from None
    try:
        import geopandas
    except ImportError:
        raise ImportError("Geopandas is required to clip voronoi regions.") from None

    if isinstance(clip, Polygon):
        clipper = geopandas.GeoDataFrame(geometry=[clip])
    elif clip is None or clip.lower() == "none":
        return regions
    elif clip.lower() in ("bounds", "bounding box", "bbox", "extent"):
        min_x, min_y, max_x, max_y = vertices.total_bounds
        bounding_poly = Polygon(
            [
                (min_x, min_y),
                (min_x, max_y),
                (max_x, max_y),
                (max_x, min_y),
                (min_x, min_y),
            ]
        )
        clipper = geopandas.GeoDataFrame(geometry=[bounding_poly])
    elif clip.lower() in ("chull", "convex hull", "convex_hull"):
        clipper = geopandas.GeoDataFrame(
            geometry=[vertices.geometry.unary_union.convex_hull]
        )
    elif clip.lower() in (
        "ahull",
        "alpha hull",
        "alpha_hull",
        "ashape",
        "alpha shape",
        "alpha_shape",
    ):
        from ..weights.distance import get_points_array
        from .alpha_shapes import alpha_shape_auto

        coordinates = get_points_array(vertices.geometry)
        clipper = geopandas.GeoDataFrame(geometry=[alpha_shape_auto(coordinates)])
    else:
        raise ValueError(
            f"Clip type '{clip}' not understood. Try one of the supported options: "
            "[None, 'extent', 'chull', 'ahull']."
        )
    clipped_regions = geopandas.overlay(regions, clipper, how="intersection")
    return clipped_regions


def voronoi_frames(
    geometry: gpd.GeoSeries | gpd.GeoDataFrame | npt.ArrayLike,
    radius: float | None = None,
    clip: str | shapely.Geometry | None = "bounding_box",
    shrink: float = 0,
    segment: float = 0,
    grid_size: float = 1e-5,
    return_input: bool | None = None,
    as_gdf: bool | None = None,
) -> gpd.GeoSeries:
    """
    Create Voronoi polygons from a GeoSeries of points, lines, or polygons.

    This is a wrapper around ``shapely.voronoi_polygons`` that handles not only
    points but also lines and polygons through their discretization and dissolution
    of the resulting polygons.

    Parameters
    ----------
    geometry : GeoSeries | GeoDataFrame | array_like
        A GeoSeries of points, lines, or polygons or an array of coordinates.
    radius : float, optional
        Deprecated. Has no effect any longer.
    clip : str, shapely.geometry.Polygon, optional
        Polygon used to clip the Voronoi polygons, by default "bounding_box"
        The options are:

        * ``None`` -- No clip is applied. Voronoi cells may be arbitrarily
          larger that the source map. Note that this may lead to cells that are many
          orders of magnitude larger in extent than the original map. Not recommended.
        * ``'bounding_box'`` -- Clip the voronoi cells to the
          bounding box of the input points.
        * ``'convex_hull'`` -- Clip the voronoi cells to the convex hull of
          the input points.
        * ``'alpha_shape'`` -- Clip the voronoi cells to the tightest hull that
          contains all points (e.g. the smallest alpha shape, using
          :func:`libpysal.cg.alpha_shape_auto`).
        * ``shapely.Polygon`` -- Clip to an arbitrary Polygon.

    shrink : float, optional
        Distance for the negative buffer of polygons required when there are polygons
        sharing portion of their exterior, by default 0
    segment : float, optional
        Distance for the segmentation of lines used to add coordinates to lines or
        polygons prior Voronoi tessellation, by default 0
    grid_size : float, optional
        Grid size precision under which the voronoi algorithm is generated,
        by default 1e-5
    return_input : bool, optional
        Whether to return the input geometry, defaults to True
    as_gdf : bool, optional
        Whether to return the output as a GeoDataFrame (True) or GeoSeries (False),
        defaults to True

    Returns
    -------
    GeoSeries | GeoDataFrame | tuple
        GeoSeries of Voronoi polygons with index allowing to link back to the input
    """
    if radius is not None:
        warnings.warn(
            "The 'radius' parameter is deprecated and will be removed in a future "
            "release. It has no effect any longer.",
            FutureWarning,
            stacklevel=2,
        )

    if isinstance(geometry, gpd.GeoDataFrame | gpd.GeoSeries):
        # Check if the input geometry is in a geographic CRS
        if geometry.crs and geometry.crs.is_geographic:
            raise ValueError(
                "Geometry is in a geographic CRS. "
                "Use 'GeoSeries.to_crs()' to re-project geometries to a "
                "projected CRS before using voronoi_polygons.",
            )
        # Set precision of the input geometry (avoids GEOS precision issues)
        objects = shapely.set_precision(geometry.geometry.copy(), grid_size)

        geom_types = objects.geom_type
        mask_poly = geom_types.isin(["Polygon", "MultiPolygon"])
        mask_line = objects.geom_type.isin(["LineString", "MultiLineString"])

        if mask_poly.any():
            # Shrink polygons if required
            if shrink != 0:
                objects[mask_poly] = objects[mask_poly].buffer(
                    -shrink, cap_style=2, join_style=2
                )
            # Segmentize polygons if required
            if segment != 0:
                objects.loc[mask_poly] = shapely.segmentize(objects[mask_poly], segment)

        if mask_line.any():
            if segment != 0:
                objects.loc[mask_line] = shapely.segmentize(objects[mask_line], segment)

            if not GPD_GE_013:
                raise ImportError(
                    "Voronoi tessellation of lines requires geopandas 0.13.0 or later."
                )

            # Remove duplicate coordinates from lines
            objects.loc[mask_line] = (
                objects.loc[mask_line]
                .get_coordinates(index_parts=True)
                .drop_duplicates(keep=False)
                .groupby(level=0)
                .apply(shapely.multipoints)
                .values
            )
    else:
        geometry = np.asarray(geometry)
        objects = geometry = gpd.GeoSeries.from_xy(geometry[:, 0], geometry[:, 1])
        mask_poly = mask_line = np.array([False])

    limit = _get_limit(objects, clip)

    # Compute Voronoi polygons
    voronoi = shapely.voronoi_polygons(
        shapely.GeometryCollection(objects.values), extend_to=limit
    )
    # Get individual polygons out of the collection
    polygons = gpd.GeoSeries(shapely.get_parts(voronoi), crs=geometry.crs)
    # Assign to each input geometry the corresponding Voronoi polygon
    # TODO: check if we still need indexing after shapely/shapely#1968 is released
    if GPD_GE_013:
        ids_objects, ids_polygons = polygons.sindex.query(
            objects, predicate="intersects"
        )
    else:
        ids_objects, ids_polygons = polygons.sindex.query_bulk(
            objects, predicate="intersects"
        )
    if mask_poly.any() or mask_line.any():
        # Dissolve polygons
        polygons = (
            polygons.iloc[ids_polygons]
            .groupby(objects.index.take(ids_objects))
            .agg(shapely.coverage_union_all)
        )
        if geometry.crs is not None:
            polygons = polygons.set_crs(geometry.crs)
    else:
        polygons = polygons.iloc[ids_polygons].reset_index(drop=True)

    # Clip polygons if limit is provided
    if limit is not None:
        to_be_clipped = polygons.sindex.query(limit.boundary, "intersects")
        polygons.iloc[to_be_clipped] = polygons.iloc[to_be_clipped].intersection(limit)

    if as_gdf is None:
        as_gdf = True
        warnings.warn(
            "The 'as_gdf' parameter currently defaults to True but will "
            "default to False in a future release. Set it explicitly to avoid "
            "this warning.",
            FutureWarning,
            stacklevel=2,
        )

    if as_gdf:
        polygons = polygons.to_frame("geometry")
        geometry = geometry.geometry.to_frame("geometry")

    if return_input is None:
        return_input = True
        warnings.warn(
            "The 'return_input' parameter currently defaults to True but will "
            "default to False in a future release. Set it explicitly to avoid "
            "this warning.",
            FutureWarning,
            stacklevel=2,
        )

    if return_input:
        return polygons, geometry

    return polygons


def _get_limit(points, clip):
    if isinstance(clip, shapely.Geometry):
        return clip
    if clip is None or clip is False:
        return None
    if clip.lower() == "none":
        warnings.warn(
            "The 'none' option for the 'clip' parameter is deprecated and will "
            "be removed in a future release. Use None or False instead.",
            FutureWarning,
            stacklevel=3,
        )
        return None
    if clip.lower() in ("bounding_box", "bounds", "bounding box", "bbox", "extent"):
        if clip.lower() != "bounding_box":
            warnings.warn(
                f"The '{clip}' option for the 'clip' parameter is deprecated and "
                "will be removed in a future release. Use 'bounding_box' instead.",
                FutureWarning,
                stacklevel=3,
            )
        return shapely.box(*points.total_bounds)

    if clip.lower() in ("chull", "convex hull", "convex_hull"):
        if clip.lower() != "convex_hull":
            warnings.warn(
                f"The '{clip}' option for the 'clip' parameter is deprecated and "
                "will be removed in a future release. Use 'convex_hull' instead.",
                FutureWarning,
                stacklevel=3,
            )
        return points.unary_union.convex_hull

    if clip.lower() in (
        "ahull",
        "alpha hull",
        "alpha_hull",
        "ashape",
        "alpha shape",
        "alpha_shape",
    ):
        if clip.lower() != "alpha_shape":
            warnings.warn(
                f"The '{clip}' option for the 'clip' parameter is deprecated and "
                "will be removed in a future release. Use 'alpha_shape' instead.",
                FutureWarning,
                stacklevel=3,
            )
        from .alpha_shapes import alpha_shape_auto

        coordinates = shapely.get_coordinates(points.values)
        return alpha_shape_auto(coordinates)

    raise ValueError(
        f"Clip type '{clip}' not understood. Try one of the supported options: "
        "[None, 'bounding_box', 'convex_hull', 'alpha_shape', shapely.Polygon]."
    )



import pytest

from .... import examples as pysal_examples
from ..wkt import WKTReader


class TesttestWKTReader:
    def setup_method(self):
        self.test_file = test_file = pysal_examples.get_path("stl_hom.wkt")
        self.obj = WKTReader(test_file, "r")

    def test_close(self):
        f = self.obj
        f.close()
        pytest.raises(ValueError, f.read)
        # w_kt_reader = WKTReader(*args, **kwargs)
        # self.assertEqual(expected, w_kt_reader.close())

    def test_open(self):
        f = self.obj
        expected = ["wkt"]
        assert expected == f.FORMATS

    def test__read(self):
        polys = self.obj.read()
        assert len(polys) == 78
        assert polys[1].centroid == (-91.195784694307383, 39.990883050220845)



import pytest


@pytest.mark.skip("skpping converters and metadata inserters")
class TestUtils:
    def test_converters(self):
        """Make a round trip to geodataframe and back."""

        raise Exception

    def test_insert_metadata(self):
        """Add an attribute to a dataframe and see if it is pervasive over copies."""

        raise Exception



# ruff: noqa: SIM115

import scipy.io as sio

from ...weights.weights import WSP, W
from .. import fileio

__author__ = "Myunghwa Hwang <mhwang4@gmail.com>"
__all__ = ["MtxIO"]


class MtxIO(fileio.FileIO):
    """
    Opens, reads, and writes weights file objects in Matrix Market ``.mtx`` format.
    The Matrix Market MTX format is used to facilitate the exchange of matrix data.
    In PySAL, it is being tested as a new file format for delivering the weights
    information of a spatial weights matrix. Although the MTX format supports both
    full and sparse matrices with different data types, it is assumed that spatial
    weights files in the ``.mtx``. format always use the sparse (or coordinate)
    format with real data values. For now, no additional assumption (e.g.,
    symmetry) is made of the structure of a weights matrix.

    With the above assumptions, the structure of a MTX file containing a spatial
    weights matrix can be defined as follows:

    ```
    %%MatrixMarket matrix coordinate real general <--- header 1 (constant)
    % Comments starts                             <---
    % ....                                           | 0 or more comment lines
    % Comments ends                               <---
    M    N    L                                   <--- header 2, rows, columns, entries
    I1   J1   A(I1,J1)                            <---
    ...                                              | L entry lines
    IL   JL   A(IL,JL)                            <---
    ```

    In the MTX format, the index for rows or columns starts with 1.

    PySAL uses ``mtx`` tools in
    `scipy.io <https://docs.scipy.org/doc/scipy/reference/tutorial/io.html>`_.
    Thus, it is subject to all limits that ``scipy`` currently has. Reengineering
    may be required, since ``scipy`` reads in the entire entry into memory.

    References
    ----------

    `MTX format specification <http://math.nist.gov/MatrixMarket/formats.html>`_

    `Matrix Market files
    <https://docs.scipy.org/doc/scipy/reference/tutorial/io.html#matrix-market-files>`_
    in ``scipy``.

    """

    FORMATS = ["mtx"]
    MODES = ["r", "w"]

    def __init__(self, *args, **kwargs):
        fileio.FileIO.__init__(self, *args, **kwargs)
        self.file = open(self.dataPath, self.mode + "b")

    def read(self, n=-1, sparse=False):  # noqa: ARG002
        """

        Parameters
        ----------
        n : int
            Read at most ``n`` objects. Default is ``-1``.
        sparse : bool
            Flag for returning a sparse weights matrix (``True``).
            Default is ``False``.

        Returns
        -------
        w : libpysal.weights.W
            A PySAL `W` object.

        """

        self._sparse = sparse
        self._complain_ifclosed(self.closed)
        return self._read()

    def seek(self, pos):
        if pos == 0:
            self.file.seek(0)
            self.pos = 0

    def _read(self):
        """Reads MatrixMarket ``.mtx`` file.

        Returns
        -------
        w : {libpysal.weights.W, libpysal.weights.WSP}
            A PySAL `W` object.

        Raises
        ------
        StopIteration
            Raised at the EOF.

        Examples
        --------

        Type ``dir(w)`` at the interpreter to see what methods are supported.
        Open a MatrixMarket ``.mtx`` file and read it into a PySAL weights object.

        >>> import libpysal
        >>> f = libpysal.io.open(libpysal.examples.get_path('wmat.mtx'), 'r')

        >>> w = f.read()

        Get the number of observations from the header.

        >>> w.n
        49

        Get the mean number of neighbors.

        >>> w.mean_neighbors
        4.73469387755102

        Get neighbor weights for a single observation.

        >>> w[1]
        {2: 0.3333, 5: 0.3333, 6: 0.3333}

        >>> f.close()

        >>> f = libpysal.io.open(libpysal.examples.get_path('wmat.mtx'), 'r')

        >>> wsp = f.read(sparse=True)

        Get the number of observations from the header.

        >>> wsp.n
        49

        Get a row from the weights matrix. Note that the first row in
        the sparse matrix (the 0th row) corresponds to ID 1 from the
        original ``.mtx`` file read in.

        >>> print(wsp.sparse[0].todense())
        [[0.     0.3333 0.     0.     0.3333 0.3333 0.     0.     0.     0.
          0.     0.     0.     0.     0.     0.     0.     0.     0.     0.
          0.     0.     0.     0.     0.     0.     0.     0.     0.     0.
          0.     0.     0.     0.     0.     0.     0.     0.     0.     0.
          0.     0.     0.     0.     0.     0.     0.     0.     0.    ]]

        """

        if self.pos > 0:
            raise StopIteration

        mtx = sio.mmread(self.file)
        # matrix market indexes start at one
        ids = list(range(1, mtx.shape[0] + 1))
        wsp = WSP(mtx, ids)

        w = wsp if self._sparse else wsp.to_W()
        self.pos += 1

        return w

    def write(self, obj):
        """Write a weights object to the opened MatrixMarket ``.mtx`` file.

        Parameters
        ----------
        obj : libpysal.weights.W
            A PySAL `W` object.

        Raises
        ------
        TypeError
            Raised when the input ``obj`` is not a PySAL `W`.

        Examples
        --------

        >>> import tempfile, libpysal, os
        >>> testfile = libpysal.io.open(libpysal.examples.get_path('wmat.mtx'), 'r')
        >>> w = testfile.read()

        Create a temporary file for this example.

        >>> f = tempfile.NamedTemporaryFile(suffix='.mtx')

        Reassign to a new variable.

        >>> fname = f.name

        Close the temporary named file.

        >>> f.close()

        Open the new file in write mode.

        >>> o = libpysal.io.open(fname, 'w')

        Write the weights object into the open file.

        >>> o.write(w)
        >>> o.close()

        Read in the newly created mtx file.

        >>> wnew = libpysal.io.open(fname, 'r').read()

        Compare values from old to new.

        >>> wnew.pct_nonzero == w.pct_nonzero
        True

        Clean up temporary file created for this example.

        >>> os.remove(fname)

        Go to the beginning of the test file.

        >>> testfile.seek(0)

        Create a sparse weights instance from the test file.

        >>> wsp = testfile.read(sparse=True)

        Open the new file in write mode.

        >>> o = libpysal.io.open(fname, 'w')

        Write the sparse weights object into the open file.

        >>> o.write(wsp)
        >>> o.close()

        Read in the newly created ``.mtx`` file.

        >>> wsp_new =  libpysal.io.open(fname, 'r').read(sparse=True)

        Compare values from old to new.

        >>> wsp_new.s0 == wsp.s0
        True

        Clean up the temporary file created for this example.

        >>> os.remove(fname)

        """

        self._complain_ifclosed(self.closed)
        if issubclass(type(obj), W) or issubclass(type(obj), WSP):
            w = obj.sparse
            sio.mmwrite(
                self.file, w, comment="Generated by PySAL", field="real", precision=7
            )
            self.pos += 1
        else:
            raise TypeError("Expected a PySAL weights object, got: %s." % (type(obj)))

    def close(self):
        self.file.close()
        fileio.FileIO.close(self)



# ruff: noqa: N811

import geodatasets
import geopandas
import numpy
import pytest
import shapely

from libpysal.graph._contiguity import _VALID_GEOMETRY_TYPES as contiguity_types
from libpysal.graph._kernel import _VALID_GEOMETRY_TYPES as kernel_types
from libpysal.graph._triangulation import _VALID_GEOMETRY_TYPES as triang_types
from libpysal.graph._utils import _validate_geometry_input

guerry = geopandas.read_file(geodatasets.get_path("geoda guerry"))
guerry["intID"] = range(len(guerry))
guerry["strID"] = guerry.intID.astype(str)
rivers = geopandas.read_file(geodatasets.get_path("eea large_rivers"))
rivers["strID"] = rivers.NAME
rivers["intID"] = rivers.index.values + 1

id_types = [None, "intID", "strID"]
id_type_names = ["no index", "int index", "string index"]

geom_names = ["guerry", "guerry centroids", "rivers"]
geoms = [guerry, guerry.set_geometry(guerry.geometry.centroid), rivers]

shuffle = [False, True]
external_ids = [False, True]
input_type = ["gdf", "gseries", "array"]

parametrize_inputs = pytest.mark.parametrize("geoms", geoms, ids=geom_names)
parametrize_idtypes = pytest.mark.parametrize("ids", id_types, ids=id_type_names)
parametrize_shuffle = pytest.mark.parametrize(
    "shuffle", shuffle, ids=["input order", "shuffled"]
)
parametrize_input_type = pytest.mark.parametrize(
    "input_type", input_type, ids=input_type
)
parametrize_external_ids = pytest.mark.parametrize(
    "external_ids", external_ids, ids=["use set_index", "use id vector"]
)


@parametrize_shuffle
@parametrize_external_ids
@parametrize_inputs
@parametrize_idtypes
@parametrize_input_type
def test_validate_input_geoms(geoms, ids, shuffle, external_ids, input_type):
    """
    Test that all combinations of geometries and ids get aligned correctly
    """
    if ids is not None:
        geoms = geoms.set_index(ids)
    input_ids = geoms.index if external_ids else None
    if shuffle:
        geoms = geoms.sample(frac=1, replace=False)
    if input_type == "gdf":
        geoms = geoms
        geom_type = geoms.geometry.iloc[0].geom_type
    elif input_type == "gseries":
        geoms = geoms.geometry
        geom_type = geoms.iloc[0].geom_type
    elif input_type == "array":
        geoms = geoms.geometry.values
        geom_type = geoms[0].geom_type
    else:
        raise ValueError(
            "input_type not in supported testing types: 'gdf', 'gseries', 'array'"
        )

    coordinates, ids, out_geoms = _validate_geometry_input(geoms, ids=input_ids)
    assert (out_geoms.index == ids).all(), "validated ids are not equal to input ids"
    if geom_type == "Point":
        assert coordinates.shape[0] == len(geoms), (
            "Point inputs should be cast to coordinates, "
            "but the output coordinates and input geometries are not equal length"
        )
        assert coordinates.shape[0] == len(ids), (
            "Point inputs should be cast to coordinates, "
            "but the output coordinates and output ids are not equal length"
        )
        if hasattr(geoms, "geometry"):
            coords = shapely.get_coordinates(geoms.geometry)
        else:
            coords = shapely.get_coordinates(geoms)
        numpy.testing.assert_array_equal(coordinates, coords)


@parametrize_shuffle
@parametrize_idtypes
def test_validate_input_coords(shuffle, ids):
    """
    Test that input coordinate arrays get validated correctly
    """
    data = guerry.sample(frac=1, replace=False) if shuffle else guerry
    input_coords = shapely.get_coordinates(data.centroid)
    if ids is not None:
        ids = data[ids].values
    coordinates, ids, out_geoms = _validate_geometry_input(input_coords, ids=ids)
    assert coordinates.shape[0] == len(out_geoms)
    assert coordinates.shape[0] == len(ids)


def test_validate_raises(
    kernel_types=kernel_types,
    contiguity_types=contiguity_types,
    triang_types=triang_types,
):
    # kernels
    with pytest.raises(ValueError):  # no lines for kernels
        _validate_geometry_input(rivers, valid_geometry_types=kernel_types)
    with pytest.raises(ValueError):  # no polygons for kernels
        _validate_geometry_input(guerry, valid_geometry_types=kernel_types)
    # triangulation
    with pytest.raises(ValueError):  # no lines for triangulation
        _validate_geometry_input(rivers, valid_geometry_types=triang_types)
    with pytest.raises(ValueError):  # no polygons for triangulation
        _validate_geometry_input(guerry, valid_geometry_types=triang_types)
    # contiguity
    with pytest.raises(ValueError):  # no point gdf for contiguity
        _validate_geometry_input(
            guerry.set_geometry(guerry.centroid),
            valid_geometry_types=contiguity_types,
        )
    with pytest.raises(ValueError):  # no point gseries for contiguity
        _validate_geometry_input(
            guerry.set_geometry(guerry.centroid).geometry,
            valid_geometry_types=contiguity_types,
        )
    with pytest.raises(ValueError):  # no point arrays for contiguity
        _validate_geometry_input(
            numpy.arange(20).reshape(-1, 2), valid_geometry_types=contiguity_types
        )


def fetch_map_string(m):
    out = m._parent.render()
    out_str = "".join(out.split())
    return out_str



"""
libpysal: Python Spatial Analysis Library (core)
================================================


Documentation
-------------
PySAL documentation is available in two forms: python docstrings and an html \
        webpage at http://pysal.org/

Available sub-packages
----------------------

cg
    Basic data structures and tools for Computational Geometry
examples
    Example data sets for testing and documentation
io
    Basic functions used by several sub-packages
weights
    Tools for creating and manipulating weights
"""

import contextlib
from importlib.metadata import PackageNotFoundError, version

from . import cg, examples, io, weights

with contextlib.suppress(PackageNotFoundError):
    __version__ = version("libpysal")
